from __future__ import annotations

import logging
from pathlib import Path
from typing import TYPE_CHECKING

import fitsio
import getdist
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy as sp
from getdist import MCSamples
from mpl_toolkits.axes_grid1 import make_axes_locatable
from vega.parameters.param_utils import build_names
from vega.plots.wedges import Wedge

from picca_bookkeeper.bookkeeper import Bookkeeper

if TYPE_CHECKING:
    from typing import Any, Dict, List, Optional, Tuple, Type

    import matplotlib

    from picca_bookkeeper.hints import Axes, Figure

logger = logging.getLogger(__name__)


def make_chain(names: List[int], mean: List[float], cov: np.ndarray) -> MCSamples:
    labels = build_names(names)
    gaussian_samples = np.random.multivariate_normal(mean, cov, size=1000000)
    samples = MCSamples(
        samples=gaussian_samples, names=names, labels=[labels[name] for name in labels]
    )
    return samples


class ReadFits:
    def __init__(
        self,
        bookkeeper: Optional[Bookkeeper | Path | str] = None,
        fit_file: Optional[Path | str] = None,
        label: Optional[str] = None,
        colour: Optional[str] = None,
        ap_baseline: Optional[float] = None,
        at_baseline: Optional[float] = None,
        blinded: bool = True,  # handle blinded(true) / unblined(false), default = blinded
    ):
        """
        Args:
            bookkeeper: Bookkeeper object to read fit information from. Could also
                be the path to a bookkeeper configuration file.
            fit_file: Vega output file (if no bookkeeper provided).
            label: Label to identify the fit run.
        """
        if bookkeeper is None:
            if fit_file is None:
                raise ValueError("Either bookkeeper or fit_file should be provided")
        elif isinstance(bookkeeper, Bookkeeper):
            self.bookkeeper = bookkeeper
        else:
            self.bookkeeper = Bookkeeper(bookkeeper)

        if fit_file is not None:
            if Path(fit_file).is_file():
                self.fit_file = Path(fit_file)
            else:
                raise FileNotFoundError(f"File does not exist, {str(fit_file)}")
        else:
            self.fit_file = self.bookkeeper.paths.fit_out_fname()

        self.label = label
        self.colour = colour

        self.ap_baseline = ap_baseline
        self.at_baseline = at_baseline

        self.read_fit()

        self.values: Dict = dict()
        self.errors: Dict = dict()
        self.covs: Dict = dict()
        self.names: List = []

    def __str__(self) -> str:
        if self.label is None:
            return self.fit_file.parents[1].name
        else:
            return self.label

    def read_fit(self) -> None:
        """
        Read relevant information from output and store it in variables.
        """
        with fitsio.FITS(self.fit_file) as hdul:
            self.values = dict()
            self.errors = dict()
            self.covs = dict()
            self.names = []
            for name, value, error, cov in zip(
                hdul["BESTFIT"]["names"].read(),
                hdul["BESTFIT"]["values"].read(),
                hdul["BESTFIT"]["errors"].read(),
                hdul["BESTFIT"]["covariance"].read(),
            ):
                self.names.append(name)
                self.values[name] = value
                self.errors[name] = error
                self.covs[name] = cov

            self.chi2 = hdul["BESTFIT"].read_header()["FVAL"]
            self.nparams = hdul["BESTFIT"].read_header()["NAXIS2"]

            self.ndata = 0
            columns = hdul["MODEL"].get_colnames()
            for column in hdul["MODEL"].get_colnames():
                if "MODEL" not in column and column[-4:] == "MASK":
                    self.ndata += hdul["MODEL"][column].read().sum()

            self.pvalue = 1 - sp.stats.chi2.cdf(self.chi2, self.ndata - self.nparams)

            self.model_header = hdul["MODEL"].read_header()

        # If fit performed in alpha/phi, translate result into ap/at.
        if ("ap" not in self.names or "at" not in self.names) and (
            "alpha" in self.names and "phi" in self.names
        ):
            self.names.append("ap")
            self.names.append("at")
            self.values["ap"] = self.values["alpha"] / np.sqrt(self.values["phi"])
            self.values["at"] = self.values["alpha"] * np.sqrt(self.values["phi"])

            self.errors["ap"] = np.sqrt(
                (self.errors["alpha"] / np.sqrt(self.values["phi"])) ** 2
                + (
                    self.errors["phi"]
                    * self.values["alpha"]
                    / (2 * self.values["phi"] ** (1.5))
                )
                ** 2
            )
            self.errors["at"] = np.sqrt(
                (self.errors["alpha"] * np.sqrt(self.values["phi"])) ** 2
                + (
                    self.errors["phi"]
                    * self.values["alpha"]
                    / np.sqrt(self.values["phi"])
                )
                ** 2
            )

        if "ap" in self.names and self.ap_baseline is not None:
            self.values["ap"] -= self.ap_baseline

        if "at" in self.names and self.at_baseline is not None:
            self.values["at"] -= self.at_baseline

    def compute_chain(self) -> None:
        res = {}
        res["chisq"] = self.chi2
        res["mean"] = list(self.values.values())
        res["cov"] = list(self.covs.values())

        res["pars"] = {
            name: {"val": self.values[name], "err": self.errors[name]}
            for name in self.names
        }

        self.chain = make_chain(res["pars"].keys(), res["mean"], res["cov"])

    @staticmethod
    def table_from_fit_data(
        fits: List[Type[ReadFits]],
        params: List[str] = ["ap", "at", "bias_LYA", "beta_LYA"],
        params_names: Optional[List[str]] = None,
        precision: int = 3,
        float_presentation: str = "f",
        blinded: bool = True,  # handle blinded(true) / unblined(false), default = blinded
    ) -> pd.DataFrame:
        if params_names is None:
            params_names = params
        else:
            params_names = params_names

        header = ["name"]
        header += params_names
        header += ["chi2 / DoF", "pvalue"]

        rows = []

        # check if blinded boolian is true / false
        if blinded:
            # Store baseline values for 'ap' and 'at' from the first fit
            baseline_ap = None
            baseline_at = None

            # First pass to determine baseline values
            for fit in fits:
                if "ap" in fit.values:
                    if baseline_ap is None:
                        baseline_ap = fit.values["ap"]
                if "at" in fit.values:
                    if baseline_at is None:
                        baseline_at = fit.values["at"]

            # Process each fit and calculate differences from the baseline
            for index, fit in enumerate(fits):
                row = []
                row.append(fit.label)

                for param in params:
                    if param in fit.values.keys():
                        if param == "ap":
                            if index == 0:
                                # For the first fit, display placeholder text with precision
                                row.append(
                                    rf"xxx ± {fit.errors[param]:.{precision}{float_presentation}}"
                                )
                            else:
                                # For subsequent fits, display the difference from the baseline
                                value_diff = fit.values[param] - baseline_ap
                                error_diff = fit.errors[
                                    param
                                ]  # Assuming the error remains the same
                                row.append(
                                    rf"{value_diff:+.{precision}{float_presentation}} ± {error_diff:.{precision}{float_presentation}}"
                                )
                        elif param == "at":
                            if index == 0:
                                # For the first fit, display placeholder text with precision
                                row.append(
                                    rf"xxx ± {fit.errors[param]:.{precision}{float_presentation}}"
                                )
                            else:
                                # For subsequent fits, display the difference from the baseline
                                value_diff = fit.values[param] - baseline_at
                                error_diff = fit.errors[
                                    param
                                ]  # Assuming the error remains the same
                                row.append(
                                    rf"{value_diff:+.{precision}{float_presentation}} ± {error_diff:.{precision}{float_presentation}}"
                                )
                        else:
                            row.append(
                                rf"{fit.values[param]:.{precision}{float_presentation}} ± {fit.errors[param]:.{precision}{float_presentation}}"
                            )
                    else:
                        row.append("")

                row.append(
                    f"{fit.chi2:.{precision}{float_presentation}}/({fit.ndata}-{fit.nparams})"
                )
                row.append(f"{fit.pvalue:.{precision}{float_presentation}}")

                rows.append(row)
        else:
            for fit in fits:
                row = []
                row.append(fit.label)

                for param in params:
                    if param in fit.values.keys():
                        if param == "ap":
                            row.append(
                                rf"{fit.values[param]:+.{precision}{float_presentation}} "
                                rf"± {fit.errors[param]:.{precision}{float_presentation}}"
                            )
                        elif param == "at":
                            row.append(
                                rf"{fit.values[param]:+.{precision}{float_presentation}} "
                                rf"± {fit.errors[param]:.{precision}{float_presentation}}"
                            )
                        else:
                            row.append(
                                rf"{fit.values[param]:.{precision}{float_presentation}} "
                                rf"± {fit.errors[param]:.{precision}{float_presentation}}"
                            )
                    else:
                        row.append("")

                row.append(
                    f"{fit.chi2:.{precision}{float_presentation}}/({fit.ndata}-{fit.nparams})"
                )
                row.append(f"{fit.pvalue:.{precision}{float_presentation}}")

                rows.append(row)

        df = pd.DataFrame(data=rows)
        df.columns = header
        # df = df.sort_values("pvalue")

        return df


class FitPlots:
    @staticmethod
    def cf_model(
        bookkeeper: Optional[Bookkeeper] = None,
        fit_file: Path | str = "",
        correlation_file: Path | str = "",
        region: str = "lya",
        region2: Optional[str] = None,
        absorber: str = "lya",
        absorber2: Optional[str] = None,
        mumin: float = 0,
        mumax: float = 1,
        ax: Optional[Axes] = None,
        r_factor: int = 2,
        plot_kwargs: Dict = dict(),
        just_return_values: bool = False,
        output_prefix: Optional[Path | str] = None,
        save_data: bool = False,
        save_dict: Dict = dict(),
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Plot fit correlation function model.

        Args:
            bookkeeper: Bookkeeper object ot use for collecting data.
            fit_file: Vega fit output file (Needed if no bookkeeper provided).
            correlation_file: Correlation output fits file (Needed if no bookkeeper
                provided).
            region: region to use.
            region2: if set use different second region.
            absorber: absorber to use.
            absorber2: if set use different second absorber.
            mumin: wedge min value.
            mumax: wedge max value.
            ax: axis where to draw the plot, if None, it'll be created.
            r_factor: exponential factor to apply to distance r.
            plot_kwargs: extra kwargs to sent to the plotting function.
            output_prefix: Save the plot under this file structure (Default: None,
                plot not saved)
            save_data: Save the data into a npz file under the output_prefix file
                structure. (Default: False).
            save_dict: Extra information to save in the npz file if save_data option
                is True. (Default: Empty dict)
        """
        if output_prefix is not None:
            output_prefix = Path(output_prefix)

        if region2 is None:
            region2 = region
        if absorber2 is None:
            absorber2 = absorber

        if fit_file != "" or correlation_file != "":
            if fit_file == "" != correlation_file == "":
                raise ValueError(
                    "Should provide fit_file and correlation_file at the same"
                    "time or use a bookkeeper"
                )

        if not isinstance(bookkeeper, Bookkeeper):
            if not Path(fit_file).is_file() or not Path(correlation_file).is_file():
                raise ValueError(
                    "Should provide fit_file and correlation_file at the same"
                    "time or use a bookkeeper"
                )

        if isinstance(bookkeeper, Bookkeeper):
            fit_file = bookkeeper.paths.fit_out_fname()
            correlation_file = bookkeeper.paths.exp_cf_fname(
                absorber, region, absorber2, region2
            )

        with fitsio.FITS(correlation_file) as ffile:
            co = ffile["COR"]["CO"][:]
            nb = ffile["COR"]["NB"][:]

            cor_header = ffile["COR"].read_header()

            N_p = cor_header["NP"]
            N_t = cor_header["NT"]

            wedge = Wedge(
                rp=(cor_header["RPMIN"], cor_header["RPMAX"], cor_header["NP"]),
                rt=(cor_header.get("RTMIN", 0), cor_header["RTMAX"], cor_header["NT"]),
                r=(cor_header.get("RTMIN", 0), cor_header["RTMAX"], cor_header["NT"]),
                mu=(mumin, mumax),
            )

        with fitsio.FITS(fit_file) as ffile:
            colnames = ffile["MODEL"].get_colnames()
            if f"{region}x{region2}_MODEL" in colnames:
                field = f"{region}x{region2}_MODEL"
            elif f"{absorber}{region}x{absorber2}{region2}_MODEL" in colnames:
                field = f"{absorber}{region}x{absorber2}{region2}_MODEL"
            else:
                raise ValueError(
                    f"Unable to find compatible card for:\n"
                    f"\tregion:{region}\n\tabsorber:{absorber}"
                    f"\tregion2:{region2}\n\tabosrber2:{absorber2}",
                    colnames,
                )

            model = np.trim_zeros(
                ffile["MODEL"][field].read(),
                "b",
            )

        if model.size != co.shape[0]:
            # What we do here is to use NT from data as the valid
            # NT for the model. Then we reshape model to remove
            # values above rp_data_max.
            logger.warning(
                "Model size is different to data size, "
                "assuming NT is the same for both"
            )

            if model.size % N_t != 0:
                raise ValueError("Unable to set NP for model.")

            model_np = model.size // N_t
            # this reshapes into data size
            model = model.reshape(model_np, N_t)[:N_p, :].reshape(-1)

        model_wedge = wedge(model, co)

        r_coef = model_wedge[0] ** r_factor

        if just_return_values:
            return (
                model_wedge[0],
                r_coef * model_wedge[1],
            )

        if save_data and output_prefix is None:
            raise ValueError("Set output_prefix in order to save data.")
        if save_data:
            data_dict = {}
        if ax is None:
            fig, ax = plt.subplots()

        ax.plot(
            model_wedge[0],
            r_coef * model_wedge[1],
            **plot_kwargs,
        )
        ax.grid(visible=True)
        ax.set_xlabel(r"$r \, [\mathrm{Mpc \, h^{-1}}]$")
        ax.set_ylabel(
            r"$r^{0} \xi(r) \, [\mathrm{{Mpc \, h^{{-1}}  }}]$".format(r_factor)
        )
        ax.set_title("{0} < $\mu$ < {1}".format(mumin, mumax))

        if save_data:
            data_dict["r"] = model_wedge[0]
            data_dict["values"] = r_coef * model_wedge[1]
            data_dict["r_factor"] = r_factor

        plt.tight_layout()
        if save_data and output_prefix is not None:
            np.savez(
                output_prefix.parent / (output_prefix.name + "-plot_cf_model.npz"),
                **{**save_dict, **data_dict},
            )

        return (model_wedge[0], r_coef * model_wedge[1])

    @staticmethod
    def xcf_model(
        bookkeeper: Optional[Bookkeeper] = None,
        fit_file: Path | str = "",
        correlation_file: Path | str = "",
        region: str = "lya",
        absorber: str = "lya",
        mumin: float = 0,
        mumax: float = 1,
        abs_mu: bool = True,
        ax: Optional[Axes] = None,
        r_factor: int = 2,
        plot_kwargs: Dict = dict(),
        just_return_values: bool = False,
        output_prefix: Optional[Path | str] = None,
        save_data: bool = False,
        save_dict: Dict = dict(),
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Plot fit correlation function model.

        Args:
            bookkeeper: Bookkeeper object ot use for collecting data.
            fit_file: Vega fit output file (Needed if no bookkeeper provided).
            correlation_file: Correlation output fits file (Needed if no bookkeeper
                provided).
            region: region to use.
            absorber: absorber to use.
            mumin: wedge min value.
            mumax: wedge max value.
            ax: axis where to draw the plot, if None, it'll be created.
            r_factor: exponential factor to apply to distance r.
            plot_kwargs: extra kwargs to sent to the plotting function.
            output_prefix: Save the plot under this file structure (Default: None,
                plot not saved)
            save_data: Save the data into a npz file under the output_prefix file
                structure. (Default: False).
            save_dict: Extra information to save in the npz file if save_data option
                is True. (Default: Empty dict)
        """
        if output_prefix is not None:
            output_prefix = Path(output_prefix)

        if fit_file != "" or correlation_file != "":
            if fit_file == "" != correlation_file == "":
                raise ValueError(
                    "Should provide fit_file and correlation_file at the same"
                    "time or use a bookkeeper"
                )

        if not isinstance(bookkeeper, Bookkeeper):
            if not Path(fit_file).is_file() or not Path(correlation_file).is_file():
                raise ValueError(
                    "Should provide fit_file and correlation_file at the same"
                    "time or use a bookkeeper"
                )

        if isinstance(bookkeeper, Bookkeeper):
            fit_file = bookkeeper.paths.fit_out_fname()
            correlation_file = bookkeeper.paths.exp_xcf_fname(absorber, region)

        with fitsio.FITS(correlation_file) as ffile:
            co = ffile["COR"]["CO"][:]
            nb = ffile["COR"]["NB"][:]

            cor_header = ffile["COR"].read_header()

            N_p = cor_header["NP"]
            N_t = cor_header["NT"]

            wedge = Wedge(
                rp=(cor_header["RPMIN"], cor_header["RPMAX"], cor_header["NP"]),
                rt=(cor_header.get("RTMIN", 0), cor_header["RTMAX"], cor_header["NT"]),
                r=(cor_header.get("RTMIN", 0), cor_header["RTMAX"], cor_header["NT"]),
                mu=(mumin, mumax),
                abs_mu=abs_mu,
            )

        with fitsio.FITS(fit_file) as ffile:
            colnames = ffile["MODEL"].get_colnames()
            for name in (
                f"qsox{region}_MODEL",
                f"qsox{absorber}{region}_MODEL",
                f"{region}xqso_MODEL",
                f"{absorber}{region}xqso_MODEL",
            ):
                if name in colnames:
                    field = name
                    break
            else:
                raise ValueError(
                    f"Unable to find compatible card for:\n"
                    f"\tregion:{region}\n\tabsorber:{absorber}",
                    colnames,
                )

            model = np.trim_zeros(
                ffile["MODEL"][field].read(),
                "b",
            )

        if model.size != co.shape[0]:
            # What we do here is to use NT from data as the valid
            # NT for the model. Then we reshape model to remove
            # values above rp_data_max.
            logger.warning(
                "Model size is different to data size, "
                "assuming NT is the same for both"
            )

            if model.size % N_t != 0:
                raise ValueError("Unable to set NP for model.")
            elif (model.size // N_t) % 2 != 0:
                raise ValueError("Unable to set NP for model. Due to odd number of NP.")

            model_np = model.size // N_t
            remove_idx = (model_np - N_p) // 2
            # this reshapes into data size
            model = model.reshape(model_np, N_t)[remove_idx:-remove_idx, :].reshape(-1)

        model_wedge = wedge(model, co)

        r_coef = model_wedge[0] ** r_factor

        if just_return_values:
            return (
                model_wedge[0],
                r_coef * model_wedge[1],
            )

        if save_data and output_prefix is None:
            raise ValueError("Set output_prefix in order to save data.")
        if save_data:
            data_dict = {}
        if ax is None:
            fig, ax = plt.subplots()

        ax.plot(
            model_wedge[0],
            r_coef * model_wedge[1],
            **plot_kwargs,
        )
        ax.grid(visible=True)
        ax.set_xlabel(r"$r \, [\mathrm{Mpc \, h^{-1}}]$")
        ax.set_ylabel(
            r"$r^{0} \xi(r) \, [\mathrm{{Mpc \, h^{{-1}}  }}]$".format(r_factor)
        )
        ax.set_title("{0} < $\mu$ < {1}".format(mumin, mumax))

        if save_data:
            data_dict["r"] = model_wedge[0]
            data_dict["values"] = r_coef * model_wedge[1]
            data_dict["r_factor"] = r_factor

        plt.tight_layout()

        if output_prefix is not None:
            if save_data:
                np.savez(
                    output_prefix.parent / (output_prefix.name + "-plot_xcf_model.npz"),
                    **{**save_dict, **data_dict},
                )
        elif save_data:
            raise ValueError("Set output_prefix in order to save data.")

        return (model_wedge[0], r_coef * model_wedge[1])

    @staticmethod
    def cf_model_map(
        bookkeeper: Optional[Bookkeeper] = None,
        fit_file: Path | str = "",
        correlation_file: Path | str = "",
        region: str = "lya",
        region2: Optional[str] = None,
        absorber: str = "lya",
        absorber2: Optional[str] = None,
        ax: Optional[Axes] = None,
        r_factor: int = 2,
        vmin: float = -0.04,
        vmax: float = 0.04,
        fig: Optional[Figure] = None,
        plot_kwargs: Dict = dict(),
        just_return_values: bool = False,
        output_prefix: Optional[Path | str] = None,
        save_data: bool = False,
        save_plot: bool = False,
        save_dict: Dict = dict(),
    ) -> Tuple[Tuple[float, ...], np.ndarray]:
        """
        Plot fit correlation function model.

        Args:
            bookkeeper: Bookkeeper object ot use for collecting data.
            fit_file: Vega fit output file (Needed if no bookkeeper provided).
            correlation_file: Correlation output fits file (Needed if no bookkeeper
                provided).
            region: region to use.
            region2: if set use different second region.
            absorber: absorber to use.
            absorber2: if set use different second absorber.
            ax: axis where to draw the plot, if None, it'll be created.
            r_factor: exponential factor to apply to distance r.
            vmin: min third axis (colormap).
            vmax: max third axis (colormap).
            plot_kwargs: extra kwargs to sent to the plotting function.
            output_prefix: Save the plot under this file structure (Default: None,
                plot not saved)
            save_data: Save the data into a npz file under the output_prefix file
                structure. (Default: False).
            save_plot: Save plot into file.
            save_dict: Extra information to save in the npz file if save_data option
                is True. (Default: Empty dict)
        """
        if output_prefix is not None:
            output_prefix = Path(output_prefix)

        if region2 is None:
            region2 = region
        if absorber2 is None:
            absorber2 = absorber

        if fit_file != "" or correlation_file != "":
            if fit_file == "" != correlation_file == "":
                raise ValueError(
                    "Should provide fit_file and correlation_file at the same"
                    "time or use a bookkeeper"
                )

        if not isinstance(bookkeeper, Bookkeeper):
            if not Path(fit_file).is_file() or not Path(correlation_file).is_file():
                raise ValueError(
                    "Should provide fit_file and correlation_file at the same"
                    "time or use a bookkeeper"
                )

        if isinstance(bookkeeper, Bookkeeper):
            fit_file = bookkeeper.paths.fit_out_fname()
            correlation_file = bookkeeper.paths.exp_cf_fname(
                absorber, region, absorber2, region2
            )

        with fitsio.FITS(correlation_file) as ffile:
            co = ffile["COR"]["CO"][:]
            rp = ffile["COR"]["RP"][:]
            rt = ffile["COR"]["RT"][:]

            cor_header = ffile["COR"].read_header()

            N_p = cor_header["NP"]
            N_t = cor_header["NT"]

        with fitsio.FITS(fit_file) as ffile:
            colnames = ffile["MODEL"].get_colnames()
            if f"{region}x{region2}_MODEL" in colnames:
                field = f"{region}x{region2}_MODEL"
            elif f"{absorber}{region}x{absorber2}{region2}_MODEL" in colnames:
                field = f"{absorber}{region}x{absorber2}{region2}_MODEL"
            else:
                raise ValueError(
                    f"Unable to find compatible card for:\n"
                    f"\tregion:{region}\n\tabsorber:{absorber}"
                    f"\tregion2:{region2}\n\tabosrber2:{absorber2}",
                    colnames,
                )

            model = np.trim_zeros(
                ffile["MODEL"][field].read(),
                "b",
            )

        if model.size != co.shape[0]:
            # What we do here is to use NT from data as the valid
            # NT for the model. Then we reshape model to remove
            # values above rp_data_max.
            logger.warning(
                "Model size is different to data size, "
                "assuming NT is the same for both"
            )

            if model.size % N_t != 0:
                raise ValueError("Unable to set NP for model.")

            model_np = model.size // N_t
            # this reshapes into data size
            model = model.reshape(model_np, N_t)[:N_p, :].reshape(-1)

        extent: tuple[float, ...]
        extent = (
            cor_header.get("RTMIN", 0.0),
            cor_header["RTMAX"],
            cor_header["RPMIN"],
            cor_header["RPMAX"],
        )
        r = np.sqrt(rp**2 + rt**2)
        nrp, nrt = cor_header["NP"], cor_header["NT"]
        r = r.reshape(nrp, nrt)

        mat = model.reshape(nrp, nrt) * r**r_factor

        if just_return_values:
            return extent, mat

        if (save_data or save_plot) and output_prefix is None:
            raise ValueError("Set output_prefix in order to save data.")
        if save_data:
            data_dict = dict(extent=extent, mat=mat)
        if ax is None:
            fig, ax = plt.subplots()
        elif fig is None:
            raise ValueError("ax and fiug should be provided at the same time")

        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)

        im = ax.imshow(
            mat,
            origin="lower",
            extent=extent,
            interpolation="nearest",
            cmap="seismic",
            **{
                **dict(vmin=vmin, vmax=vmax),
                **plot_kwargs,
            },
        )

        fig.colorbar(im, cax=cax, orientation="vertical")
        ax.set_xlabel(r"$r_{\perp} \, [h^{-1} \, \rm{Mpc}]$")
        ax.set_ylabel(r"$r_{\parallel} \, [h^{-1} \, \rm{Mpc}]$")
        cax.yaxis.set_label_position("right")
        cax.set_ylabel(r"$r\xi(r_{\parallel,r_{\perp}})$")

        if output_prefix is not None:
            if save_plot:
                output_prefix = Path(output_prefix)
                plt.tight_layout()
                plt.savefig(
                    str(output_prefix) + ".png",
                    dpi=300,
                )

            if save_data:
                np.savez(
                    str(output_prefix) + ".npz",
                    **{**save_dict, **data_dict},
                )
        elif save_data or save_plot:
            raise ValueError("Set output_prefix in order to save data.")

        return extent, mat

    @staticmethod
    def rp_model(
        bookkeeper: Optional[Bookkeeper] = None,
        fit_file: Path | str = "",
        correlation_file: Path | str = "",
        auto: Optional[bool] = False,
        region: str = "lya",
        region2: Optional[str] = None,
        absorber: str = "lya",
        absorber2: Optional[str] = None,
        rtmin: float = 0,
        rtmax: float = 1,
        ax: Optional[Axes] = None,
        r_factor: int = 2,
        plot_kwargs: Dict = dict(),
        just_return_values: bool = False,
        output_prefix: Optional[Path | str] = None,
        save_data: bool = False,
        save_dict: Dict = dict(),
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Plot fit correlation function model.

        Args:
            bookkeeper: Bookkeeper object ot use for collecting data.
            fit_file: Vega fit output file (Needed if no bookkeeper provided).
            correlation_file: Correlation output fits file (Needed if no bookkeeper
                provided).
            auto. Use auto-correlation
            region: region to use.
            region2: if set use different second region.
            absorber: absorber to use.
            absorber2: if set use different second absorber.
            mumin: wedge min value.
            mumax: wedge max value.
            ax: axis where to draw the plot, if None, it'll be created.
            r_factor: exponential factor to apply to distance r.
            plot_kwargs: extra kwargs to sent to the plotting function.
            output_prefix: Save the plot under this file structure (Default: None,
                plot not saved)
            save_data: Save the data into a npz file under the output_prefix file
                structure. (Default: False).
            save_dict: Extra information to save in the npz file if save_data option
                is True. (Default: Empty dict)
        """
        if output_prefix is not None:
            output_prefix = Path(output_prefix)

        if region2 is None:
            region2 = region
        if absorber2 is None:
            absorber2 = absorber

        if fit_file != "" or correlation_file != "":
            if fit_file == "" != correlation_file == "":
                raise ValueError(
                    "Should provide fit_file and correlation_file at the same"
                    "time or use a bookkeeper"
                )

        if not isinstance(bookkeeper, Bookkeeper):
            if not Path(fit_file).is_file() or not Path(correlation_file).is_file():
                raise ValueError(
                    "Should provide fit_file and correlation_file at the same"
                    "time or use a bookkeeper"
                )

        if isinstance(bookkeeper, Bookkeeper):
            fit_file = bookkeeper.paths.fit_out_fname()
            if auto:
                correlation_file = bookkeeper.paths.exp_cf_fname(
                    absorber, region, absorber2, region2
                )
                name = "cf_rp"
            else:
                correlation_file = bookkeeper.paths.exp_xcf_fname(
                    absorber,
                    region,
                )
                name = "xcf_rp"

        with fitsio.FITS(correlation_file) as ffile:
            co = ffile["COR"]["CO"][:]
            nb = ffile["COR"]["NB"][:]

            cor_header = ffile["COR"].read_header()

            N_p = cor_header["NP"]
            N_t = cor_header["NT"]
            rp = ffile["COR"]["RP"][:]
            rt = ffile["COR"]["RT"][:]
            co = ffile["COR"]["CO"][:]

            cor_header = ffile["COR"].read_header()

        weights = 1 / np.diag(co).reshape(N_p, N_t)
        rt = rt.reshape(N_p, N_t)
        rp = rp.reshape(N_p, N_t)

        w = (rt >= rtmin) & (rt <= rtmax)
        weights[~w] = 0

        with fitsio.FITS(fit_file) as ffile:
            colnames = ffile["MODEL"].get_colnames()
            if auto:
                if f"{region}x{region2}_MODEL" in colnames:
                    field = f"{region}x{region2}_MODEL"
                elif f"{absorber}{region}x{absorber2}{region2}_MODEL" in colnames:
                    field = f"{absorber}{region}x{absorber2}{region2}_MODEL"
                else:
                    raise ValueError(
                        f"Unable to find compatible card for:\n"
                        f"\tregion:{region}\n\tabsorber:{absorber}"
                        f"\tregion2:{region2}\n\absorber2:{absorber2}",
                        colnames,
                    )
            else:
                for name in (
                    f"qsox{region}_MODEL",
                    f"qsox{absorber}{region}_MODEL",
                    f"{region}xqso_MODEL",
                    f"{absorber}{region}xqso_MODEL",
                ):
                    if name in colnames:
                        field = name
                        break
                else:
                    raise ValueError(
                        f"Unable to find compatible card for:\n"
                        f"\tregion:{region}\n\tabsorber:{absorber}"
                        f"\tregion2:{region2}\n\absorber2:{absorber2}",
                        colnames,
                    )

            model = np.trim_zeros(
                ffile["MODEL"][field].read(),
                "b",
            )

        if model.size != co.shape[0]:
            # What we do here is to use NT from data as the valid
            # NT for the model. Then we reshape model to remove
            # values above rp_data_max.
            logger.warning(
                "Model size is different to data size, "
                "assuming NT is the same for both"
            )

            if model.size % N_t != 0:
                raise ValueError("Unable to set NP for model.")

            model_np = model.size // N_t
            if auto:
                # this reshapes into data size
                model = model.reshape(model_np, N_t)[:N_p, :].reshape(-1)

            else:
                remove_idx = (model_np - N_p) // 2
                # this reshapes into data size
                model = model.reshape(model_np, N_t)[remove_idx:-remove_idx, :].reshape(
                    -1
                )

        mat = model.reshape(N_p, N_t)
        data = np.average(mat, weights=weights, axis=1)
        rp = np.average(rp, weights=weights, axis=1)

        if just_return_values:
            return (
                rp,
                data,
            )

        if save_data and output_prefix is None:
            raise ValueError("Set output_prefix in order to save data.")
        if save_data:
            data_dict = {}
        if ax is None:
            fig, ax = plt.subplots()

        ax.plot(
            rp,
            data,
            **plot_kwargs,
        )
        ax.grid(visible=True)
        ax.set_xlabel(r"$r_\parallel  \, [\mathrm{{Mpc \, h^{{-1}}  }}]$")
        ax.set_ylabel(r"$\xi(r) \, [\mathrm{{Mpc \, h^{{-1}}  }}]$")
        ax.set_title(
            rf"${rtmin} < r_\perp < {rtmax} \, \, [\mathrm{{Mpc \, h^{{-1}}}}]$"
        )

        if save_data:
            data_dict["rp"] = rp
            data_dict["data"] = data

        plt.tight_layout()
        if save_data and output_prefix is not None:
            np.savez(
                output_prefix.parent / (output_prefix.name + f"-plot_{name}_model.npz"),
                **{**save_dict, **data_dict},
            )

        return (rp, data)

    @staticmethod
    def xcf_model_map(
        bookkeeper: Optional[Bookkeeper] = None,
        fit_file: Path | str = "",
        correlation_file: Path | str = "",
        region: str = "lya",
        absorber: str = "lya",
        ax: Optional[Axes] = None,
        r_factor: int = 2,
        vmin: float = -0.4,
        vmax: float = 0.4,
        fig: Optional[Figure] = None,
        plot_kwargs: Dict = dict(),
        just_return_values: bool = False,
        output_prefix: Optional[Path | str] = None,
        save_data: bool = False,
        save_plot: bool = False,
        save_dict: Dict = dict(),
    ) -> Tuple[Tuple[float, ...], np.ndarray]:
        """
        Plot fit correlation function model.

        Args:
            bookkeeper: Bookkeeper object ot use for collecting data.
            fit_file: Vega fit output file (Needed if no bookkeeper provided).
            correlation_file: Correlation output fits file (Needed if no bookkeeper
                provided).
            region: region to use.
            absorber: absorber to use.
            ax: axis where to draw the plot, if None, it'll be created.
            r_factor: exponential factor to apply to distance r.
            vmin: min third axis for the colormap.
            vmax: max third axis for the colormap.
            plot_kwargs: extra kwargs to sent to the plotting function.
            output_prefix: Save the plot under this file structure (Default: None,
                plot not saved)
            save_data: Save the data into a npz file under the output_prefix file
                structure. (Default: False).
            save_plot: Save the plot into a png file.
            save_dict: Extra information to save in the npz file if save_data option
                is True. (Default: Empty dict)
        """
        if output_prefix is not None:
            output_prefix = Path(output_prefix)

        if fit_file != "" or correlation_file != "":
            if fit_file == "" != correlation_file == "":
                raise ValueError(
                    "Should provide fit_file and correlation_file at the same"
                    "time or use a bookkeeper"
                )

        if not isinstance(bookkeeper, Bookkeeper):
            if not Path(fit_file).is_file() or not Path(correlation_file).is_file():
                raise ValueError(
                    "Should provide fit_file and correlation_file at the same"
                    "time or use a bookkeeper"
                )

        if isinstance(bookkeeper, Bookkeeper):
            fit_file = bookkeeper.paths.fit_out_fname()
            correlation_file = bookkeeper.paths.exp_xcf_fname(absorber, region)

        with fitsio.FITS(correlation_file) as ffile:
            co = ffile["COR"]["CO"][:]
            nb = ffile["COR"]["NB"][:]
            rp = ffile["COR"]["RP"][:]
            rt = ffile["COR"]["RT"][:]

            cor_header = ffile["COR"].read_header()

            N_p = cor_header["NP"]
            N_t = cor_header["NT"]

        with fitsio.FITS(fit_file) as ffile:
            colnames = ffile["MODEL"].get_colnames()
            for name in (
                f"qsox{region}_MODEL",
                f"qsox{absorber}{region}_MODEL",
                f"{region}xqso_MODEL",
                f"{absorber}{region}xqso_MODEL",
            ):
                if name in colnames:
                    field = name
                    break
            else:
                raise ValueError(
                    f"Unable to find compatible card for:\n"
                    f"\tregion:{region}\n\tabsorber:{absorber}",
                    colnames,
                )

            model = np.trim_zeros(
                ffile["MODEL"][field].read(),
                "b",
            )

        if model.size != co.shape[0]:
            # What we do here is to use NT from data as the valid
            # NT for the model. Then we reshape model to remove
            # values above rp_data_max.
            logger.warning(
                "Model size is different to data size, "
                "assuming NT is the same for both"
            )

            if model.size % N_t != 0:
                raise ValueError("Unable to set NP for model.")
            elif (model.size // N_t) % 2 != 0:
                raise ValueError("Unable to set NP for model. Due to odd number of NP.")

            model_np = model.size // N_t
            remove_idx = (model_np - N_p) // 2
            # this reshapes into data size
            model = model.reshape(model_np, N_t)[remove_idx:-remove_idx, :].reshape(-1)

        extent: tuple[float, ...]
        extent = (
            cor_header.get("RTMIN", 0),
            cor_header["RTMAX"],
            cor_header["RPMIN"],
            cor_header["RPMAX"],
        )
        r = np.sqrt(rp**2 + rt**2)
        nrp, nrt = cor_header["NP"], cor_header["NT"]
        r = r.reshape(nrp, nrt)

        mat = model.reshape(nrp, nrt) * r**r_factor

        if just_return_values:
            return extent, mat

        if (save_data or save_plot) and output_prefix is None:
            raise ValueError("Set output_prefix in order to save data.")
        if save_data:
            data_dict = dict(extent=extent, mat=mat)
        if ax is None:
            fig, ax = plt.subplots()
        elif fig is None:
            raise ValueError("ax and fiug should be provided at the same time")

        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)

        im = ax.imshow(
            mat,
            origin="lower",
            extent=extent,
            interpolation="nearest",
            cmap="seismic",
            **{
                **dict(vmin=vmin, vmax=vmax),
                **plot_kwargs,
            },
        )

        fig.colorbar(im, cax=cax, orientation="vertical")
        ax.set_xlabel(r"$r_{\perp} \, [h^{-1} \, \rm{Mpc}]$")
        ax.set_ylabel(r"$r_{\parallel} \, [h^{-1} \, \rm{Mpc}]$")
        cax.yaxis.set_label_position("right")
        cax.set_ylabel(r"$r\xi(r_{\parallel,r_{\perp}})$")

        if output_prefix is not None:
            if save_plot:
                output_prefix = Path(output_prefix)
                plt.tight_layout()
                plt.savefig(
                    str(output_prefix) + ".png",
                    dpi=300,
                )

            if save_data:
                np.savez(
                    str(output_prefix) + ".npz",
                    **{**save_dict, **data_dict},
                )
        elif save_data or save_plot:
            raise ValueError("Set output_prefix in order to save data.")

        return extent, mat

    @staticmethod
    def plot_errorbars_from_fit(
        readfits: List[ReadFits],
        param: str,
        param_name: Optional[str] = None,
        ax: Optional[Axes] = None,
        plot_kwargs: Dict = dict(),
        reference: Optional[ReadFits] = None,
        blinded: bool = True,  # Add the blinded parameter, default is True
    ) -> List[matplotlib.container.Container | Any]:
        """
        Args:
            readfits: List of readfits objects to show in the plot.
            param: Param to plot.
            param_name: Name of the param to show.
            ax: Axis where to plot. If not provided, it will be created.
            reference: Add reference as shaded line.
            blinded: Whether to use the blinded or unblinded version of the function.

        Returns:
            List of plot handles to make legend from it.
        """
        if param_name is None:
            param_name = param

        if ax is None:
            fig, ax = plt.subplots()

        handles: List[Any] = []

        # Check if blinded boolean is true / false
        if blinded:
            # Blinded code block

            # Baseline setup
            baseline_value = None
            baseline_error = None

            if param in ("ap", "at") and reference:
                # Set the baseline using the first fit in the list
                baseline_value = reference.values.get(param, None)
                baseline_error = reference.errors.get(
                    param, 0
                )  # Assume first fit error is baseline error

                if baseline_value is None:
                    baseline_value = reference.model_header.get(param, None)
                if baseline_value is None and reference is not None:
                    baseline_value = reference.values.get(param, None)
                if baseline_value is None:
                    baseline_value = 0

                # Plot the shaded regions for ap/at, centered at 0
                ax.axvspan(
                    -baseline_error,
                    baseline_error,
                    alpha=0.2,
                    color="gray",
                )
                ax.axvspan(
                    -baseline_error / 3,
                    baseline_error / 3,
                    alpha=0.2,
                    color="red",
                )

                # Plot the baseline vertical line at 0
                handles.append(
                    ax.axvline(
                        0,  # Baseline is represented as 0
                        color="black",
                        ls="--",
                        lw=0.6,
                        alpha=1,
                        label="Baseline",
                    )
                )

            # Plot the error bars for each fit
            for i, fit in enumerate(readfits):
                value = fit.values.get(param, None)
                error = fit.errors.get(param, 0)

                if value is None:
                    value = fit.model_header.get(param, None)
                    if value is None and reference is not None:
                        value = reference.values.get(param, None)
                    error = 0

                if param in ("ap", "at") and baseline_value is not None:
                    value -= baseline_value  # Adjust value to be the difference from baseline
                    # Error remains the same

                if fit.colour is not None:
                    plot_kwargs = {**plot_kwargs, **dict(color=fit.colour)}

                # Plot the errorbar with adjusted value (difference from baseline for ap/at)
                handles.append(
                    ax.errorbar(
                        value,
                        i,
                        xerr=error,
                        yerr=0,
                        label=fit.label,
                        marker="o",
                        **plot_kwargs,
                    )
                )

            # Adjust x-axis label for ap/at
            if param == "ap":
                ax.set_xlabel("delta_ap")
            elif param == "at":
                ax.set_xlabel("delta_at")
            else:
                ax.set_xlabel(param_name)

            # Add shaded regions to the 3rd, 4th, and 5th plots
            if reference is not None and param not in ("ap", "at"):
                if param in reference.values.keys():
                    value = reference.values.get(param, None)
                    error = reference.errors.get(param, 0)
                    ax.axvspan(
                        value - error,
                        value + error,
                        alpha=0.2,
                        color="gray",
                    )
                    handles.append(
                        ax.axvline(
                            value,
                            color="black",
                            ls="--",
                            lw=0.6,
                            alpha=1,
                            label=reference.label,
                        )
                    )

            ax.set_yticks(range(len(readfits)))  # Label each fit
            ax.grid(visible=True)

        else:
            # Unblinded code block (this runs when blinded=False)
            if reference is not None and param in reference.values.keys():
                value = reference.values.get(param, None)
                error = reference.errors.get(param, 0)
                ax.axvspan(
                    value - error,
                    value + error,
                    alpha=0.2,
                    color="gray",
                )
                if param in ("ap", "at"):
                    ax.axvspan(
                        value - error,
                        value - error / 3,
                        alpha=0.2,
                        color="red",
                    )
                    ax.axvspan(
                        value + error / 3,
                        value + error,
                        alpha=0.2,
                        color="red",
                    )
                handles.append(
                    ax.axvline(
                        value,
                        color="black",
                        ls="--",
                        lw=0.6,
                        alpha=1,
                        label=reference.label,
                    )
                )

            for i, fit in enumerate(readfits):
                value = fit.values.get(param, None)
                error = fit.errors.get(param, 0)

                if value is None:
                    value = fit.model_header.get(param, None)

                    if value is None and reference is not None:
                        value = reference.values.get(param, None)
                    error = 0

                if fit.colour is not None:
                    plot_kwargs = {**plot_kwargs, **dict(color=fit.colour)}

                handles.append(
                    ax.errorbar(
                        value,
                        i,
                        xerr=error,
                        yerr=0,
                        label=fit.label,
                        marker="o",
                        **plot_kwargs,
                    )
                )

            ax.set_xlabel(param_name)
            ax.set_yticks([])  # No ticks on y-axis for unblinded case
            ax.grid(visible=True)

        return handles

    @staticmethod
    def plot_p_value_from_fit(
        readfits: List[ReadFits],
        ax: Optional[Axes] = None,
        plot_kwargs: Dict = dict(),
        reference: Optional[ReadFits] = None,
    ) -> List[matplotlib.container.Container | Any]:
        """
        Args:
            readfits: List of readfits objects to show in the plot.
            ax: Axis where to plot. If not provided, it will be created.

        Returns:
            List of plot handles to make legend from it.
        """
        if ax is None:
            fig, ax = plt.subplots()

        handles: List[Any] = []

        if reference is not None:
            handles.append(
                ax.axvline(
                    reference.pvalue,
                    color="black",
                    ls="--",
                    lw=0.6,
                    alpha=1,
                    label=reference.label,
                )
            )

        for i, fit in enumerate(readfits):
            if fit.colour is not None:
                plot_kwargs = {**plot_kwargs, **dict(color=fit.colour)}

            handles.append(
                ax.errorbar(
                    fit.pvalue / 2,
                    i,
                    xerr=fit.pvalue / 2,
                    yerr=0,
                    label=fit.label,
                    **plot_kwargs,
                )
            )

        ax.set_xlabel("p-value")
        ax.set_yticks([])
        ax.grid(visible=True)

        return handles

    @staticmethod
    def plot_triangle(
        param_names: List[str],
        readfits: Optional[List[ReadFits]] = None,
        chains: Optional[List[MCSamples]] = None,
        labels: Optional[List[str | None]] = None,
        colours: Optional[List[str | None]] = None,
        g: Optional[getdist.plots.GetDistPlotter] = None,
        plot_kwargs: Dict = dict(),
    ) -> None:
        if readfits is None and chains is None:
            raise ValueError("Either provide chains or ReadFits objects.")

        if readfits is not None:
            chains = []
            for fit in readfits:
                if hasattr(fit, "chain"):
                    chains.append(fit.chain)
                else:
                    raise AttributeError(
                        "Compute chain for all fits by using the method "
                        "compute_chain before running the triangle plot."
                    )

            if labels is None:
                labels = [fit.label for fit in readfits]
            if colours is None:
                colours = [fit.colour for fit in readfits]

        if g is None:
            g = getdist.plots.getSubplotPlotter()

        g.triangle_plot(
            chains,
            param_names,
            legend_labels=labels,
            contour_colors=colours,
            contours_ls=["solid", "dashed", "dotted"],
            **plot_kwargs,
        )



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

import fitsio
import matplotlib.pyplot as plt
import numpy as np
from mpl_toolkits.axes_grid1 import make_axes_locatable
from vega.plots.wedges import Wedge

from picca_bookkeeper.bookkeeper import Bookkeeper

if TYPE_CHECKING:
    from typing import Dict, Optional, Tuple

    from picca_bookkeeper.hints import Axes, Figure


class CorrelationPlots:
    @staticmethod
    def cf(
        bookkeeper: Optional[Bookkeeper],
        region: str = "lya",
        region2: Optional[str] = None,
        absorber: str = "lya",
        absorber2: Optional[str] = None,
        correlation_file: Path | str = "",
        mumin: float = 0,
        mumax: float = 1,
        ax: Optional[Axes] = None,
        r_factor: int = 2,
        plot_kwargs: Dict = dict(),
        just_return_values: bool = False,
        output_prefix: Optional[Path | str] = None,
        save_data: bool = False,
        save_plot: bool = False,
        save_dict: Dict = dict(),
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """
        Plotting correlation function in defined wedge.

        Arguments:
        ----------
        bookkeeper: bookkeeper object to use for collect data.
        region: region to use.
        region2: if set use different second region.
        absorber: absorber to use.
        absorber2: if set use different second absorber.
        correlation_file: Provide path to correlation file alternatively.
        mumin: wedge min value.
        mumax: wedge max value.
        ax: axis where to draw the plot, if None, it'll be created.
        r_factor: exponential factor to apply to distance r.
        plot_kwargs: extra kwargs to sent to the plotting function.
        output_prefix: Save the plot under this file structure (Default: None, plot not saved)
        save_data: Save the data into a npz file under the output_prefix file structure. (Default: False).
        save_plot: Save the plot into a png file. (Default: False).
        save_dict: Extra information to save in the npz file if save_data option is True. (Default: Empty dict)
        """
        if output_prefix is not None:
            output_prefix = Path(output_prefix)

        if isinstance(bookkeeper, Bookkeeper):
            correlation_file = bookkeeper.paths.exp_cf_fname(
                absorber, region, absorber2, region2
            )

        with fitsio.FITS(correlation_file) as ffile:
            try:
                da = ffile["COR"]["DA"][:]
            except ValueError:
                da = ffile["COR"]["DA_BLIND"][:]
            co = ffile["COR"]["CO"][:]
            nb = ffile["COR"]["NB"][:]

            cor_header = ffile["COR"].read_header()
            wedge = Wedge(
                rp=(cor_header["RPMIN"], cor_header["RPMAX"], cor_header["NP"]),
                rt=(cor_header.get("RTMIN", 0), cor_header["RTMAX"], cor_header["NT"]),
                r=(cor_header.get("RTMIN", 0), cor_header["RTMAX"], cor_header["NT"]),
                mu=(mumin, mumax),
            )
        data_wedge = wedge(da, co)

        # Computing nb wedge manually
        axis_sum = wedge.weights.sum(axis=0)
        w = axis_sum > 0
        nb_weights = np.copy(wedge.weights)
        nb_weights[:, w] /= wedge.weights.sum(axis=0)[w]
        nb_wedge = data_wedge[0], nb_weights.dot(nb)

        r_coef = data_wedge[0] ** r_factor

        if just_return_values:
            return (
                data_wedge[0],
                r_coef * data_wedge[1],
                r_coef * np.sqrt(np.diag(data_wedge[2])),
                nb_wedge[1],
            )

        if (save_data or save_plot) and output_prefix is None:
            raise ValueError("Set output_prefix in order to save data.")
        if save_data:
            data_dict = {}
        if ax is None:
            fig, ax = plt.subplots()

        ax.errorbar(
            data_wedge[0],
            r_coef * data_wedge[1],
            yerr=r_coef * np.sqrt(np.diag(data_wedge[2])),
            **plot_kwargs,
        )
        ax.grid(visible=True)
        ax.set_xlabel(r"$r \, [\mathrm{Mpc \, h^{-1}}]$")
        ax.set_ylabel(
            r"$r^{0} \xi(r) \, [\mathrm{{Mpc \, h^{{-1}}  }}]$".format(r_factor)
        )
        ax.set_title("{0} < $\mu$ < {1}".format(mumin, mumax))

        if save_data:
            data_dict["r"] = data_wedge[0]
            data_dict["values"] = r_coef * data_wedge[1]
            data_dict["errors"] = r_coef * np.sqrt(np.diag(data_wedge[2]))
            data_dict["r_factor"] = r_factor
            data_dict["nb"] = nb_wedge[1]
            # data_dict['wedge_data'] = data_wedge

        plt.tight_layout()

        if output_prefix is not None:
            if save_plot:
                plt.savefig(
                    output_prefix.parent / (output_prefix.name + "-plot_cf.png"),
                )
            if save_data:
                np.savez(
                    output_prefix.parent / (output_prefix.name + "-plot_cf.npz"),
                    **{**save_dict, **data_dict},
                )
        elif save_data or save_plot:
            raise ValueError("Set output_prefix in order to save data.")

        return (
            data_wedge[0],
            r_coef * data_wedge[1],
            r_coef * np.sqrt(np.diag(data_wedge[2])),
            nb_wedge[1],
        )

    @staticmethod
    def rp(
        bookkeeper: Optional[Bookkeeper],
        region: str = "lya",
        auto: Optional[bool] = False,
        region2: Optional[str] = None,
        absorber: str = "lya",
        absorber2: Optional[str] = None,
        correlation_file: Path | str = "",
        rtmin: float = 0,
        rtmax: float = 4,
        rebin: Optional[int] = None,
        ax: Optional[Axes] = None,
        plot_kwargs: Dict = dict(),
        just_return_values: bool = False,
        output_prefix: Optional[Path | str] = None,
        save_data: bool = False,
        save_plot: bool = False,
        save_dict: Dict = dict(),
        tracer: str = "qso",
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """
        Plotting correlation function in defined wedge.

        Arguments:
        ----------
        bookkeeper: bookkeeper object to use for collect data.
        region: region to use.
        auto: Show auto-correlation (not cross).
        region2: if set use different second region.
        absorber: absorber to use.
        absorber2: if set use different second absorber.
        correlation_file: Provide path to correlation file alternatively.
        rtmin: in Mpc/h.
        rtmax: in Mpc/h
        ax: axis where to draw the plot, if None, it'll be created.
        plot_kwargs: extra kwargs to sent to the plotting function.
        output_prefix: Save the plot under this file structure (Default: None, plot not saved)
        save_data: Save the data into a npz file under the output_prefix file structure. (Default: False).
        save_plot: Save the plot into a png file. (Default: False).
        save_dict: Extra information to save in the npz file if save_data option is True. (Default: Empty dict)
        tracer: Tracer to use.
        """
        if output_prefix is not None:
            output_prefix = Path(output_prefix)

        if isinstance(bookkeeper, Bookkeeper):
            if auto:
                correlation_file = bookkeeper.paths.exp_cf_fname(
                    absorber, region, absorber2, region2
                )
                name = "cf_rp"
            else:
                correlation_file = bookkeeper.paths.exp_xcf_fname(
                    absorber, region, tracer
                )
                name = "xcf_rp"

        with fitsio.FITS(correlation_file) as ffile:
            try:
                da = ffile["COR"]["DA"][:]
            except ValueError:
                da = ffile["COR"]["DA_BLIND"][:]
            co = ffile["COR"]["CO"][:]
            nb = ffile["COR"]["NB"][:]
            rp = ffile["COR"]["RP"][:]
            rt = ffile["COR"]["RT"][:]

            cor_header = ffile["COR"].read_header()
            nrp, nrt = cor_header["NP"], cor_header["NT"]

        mat = da.reshape(nrp, nrt)
        errmat = np.sqrt(np.diag(co)).reshape(nrp, nrt)
        nmat = nb.reshape(nrp, nrt)

        weights = np.copy(1 / errmat**2)  # 1 / errmat**2

        rt = rt.reshape(nrp, nrt)
        rp = rp.reshape(nrp, nrt)

        w = (rt >= rtmin) & (rt <= rtmax)
        weights[~w] = 0

        if rebin is not None:
            if nrp % rebin != 0:
                raise ValueError("NP should be divisible by rebin, NP: ", nrt)

            mat = mat.reshape(nrp // rebin, -1)
            errmat = errmat.reshape(nrp // rebin, -1)
            nmat = nmat.reshape(nrp // rebin, -1)
            rp = rp.reshape(nrp // rebin, -1)
            weights = weights.reshape(nrp // rebin, -1)

        rp = np.average(rp, weights=weights, axis=1)
        data = np.average(mat, weights=weights, axis=1)
        error = np.sqrt(np.average(errmat**2, weights=weights**2, axis=1))
        nb = np.sum(nmat, axis=1)

        if just_return_values:
            return (rp, data, error, nb)

        if (save_data or save_plot) and output_prefix is None:
            raise ValueError("Set output_prefix in order to save data.")
        if save_data:
            data_dict = {}
        if ax is None:
            fig, ax = plt.subplots()

        ax.errorbar(
            rp,
            data,
            yerr=error,
            **plot_kwargs,
        )
        ax.grid(visible=True)
        ax.set_title(
            rf"${rtmin} < r_\perp < {rtmax} \, \, [\mathrm{{Mpc \, h^{{-1}}}}]$"
        )
        ax.set_xlabel(r"$r_\parallel  \, [\mathrm{{Mpc \, h^{{-1}}  }}]$")
        ax.set_ylabel(r"$\xi(r) \, [\mathrm{{Mpc \, h^{{-1}}  }}]$")
        if save_data:
            data_dict["rp"] = rp
            data_dict["data"] = data
            data_dict["errors"] = error
            data_dict["nb"] = nb
            # data_dict['wedge_data'] = data_wedge

        plt.tight_layout()

        if output_prefix is not None:
            if save_plot:
                plt.savefig(
                    output_prefix.parent / (output_prefix.name + f"-plot_{name}.png"),
                )
            if save_data:
                np.savez(
                    output_prefix.parent / (output_prefix.name + f"-plot_{name}.npz"),
                    **{**save_dict, **data_dict},
                )
        elif save_data or save_plot:
            raise ValueError("Set output_prefix in order to save data.")

        return (rp, data, error, nb)

    @staticmethod
    def compare_cfs(
        bkp1: Bookkeeper,
        bkp2: Optional[Bookkeeper] = None,
        mumin: float = 0,
        mumax: float = 1,
        region: str = "lya",
        region2: Optional[str] = None,
        correlation_file: Path | str = "",
        correlation_file2: Path | str = "",
        label: Optional[str] = None,
        label2: Optional[str] = None,
    ) -> None:
        fig, axs = plt.subplots(
            3, 1, sharex=True, gridspec_kw={"height_ratios": [3, 2, 2]}
        )

        if label is None and isinstance(bkp1, Bookkeeper):
            label = bkp1.label
        if label2 is None and isinstance(bkp2, Bookkeeper):
            label2 = bkp2.label

        ax = axs[0]
        a = CorrelationPlots.cf(
            bkp1,
            region=region,
            region2=region2,
            r_factor=2,
            ax=ax,
            correlation_file=correlation_file,
            mumin=mumin,
            mumax=mumax,
            plot_kwargs=dict(label=label),
        )
        b = CorrelationPlots.cf(
            bkp2,
            region=region,
            region2=region2,
            correlation_file=correlation_file2,
            r_factor=2,
            ax=ax,
            mumin=mumin,
            mumax=mumax,
            plot_kwargs=dict(label=label2),
        )
        ax.grid(visible=True)
        #    ax.set_ylim(-1, 1)
        ax.legend()

        ax = axs[1]
        ax.plot(a[0], b[1] - a[1], c="k")
        ax.set_xlabel(axs[0].get_xlabel())
        ax.set_ylabel(f"{label2}-{label}")
        ax.grid(visible=True)
        ax.set_xlabel(None)

        ax = axs[2]
        ax.plot(a[0], b[2] / a[2], c="k")
        ax.set_xlabel(axs[0].get_xlabel())
        ax.set_ylabel(f"errors{label2}/{label}")
        ax.grid(visible=True)

        # axs[1].set_ylim(0.85, 1.15)
        axs[0].set_xlabel(None)

    @staticmethod
    def compare_xcfs(
        bkp1: Bookkeeper,
        bkp2: Optional[Bookkeeper] = None,
        mumin: float = 0,
        mumax: float = 1,
        region: str = "lya",
        correlation_file: str = "",
        correlation_file2: str = "",
        label: Optional[str] = None,
        label2: Optional[str] = None,
    ) -> None:
        fig, axs = plt.subplots(
            3, 1, sharex=True, gridspec_kw={"height_ratios": [3, 2, 2]}
        )

        if label is None and isinstance(bkp1, Bookkeeper):
            label = bkp1.label
        if label2 is None and isinstance(bkp2, Bookkeeper):
            label2 = bkp2.label

        ax = axs[0]
        a = CorrelationPlots.xcf(
            bkp1,
            region=region,
            r_factor=2,
            ax=ax,
            correlation_file=correlation_file,
            mumin=mumin,
            mumax=mumax,
            plot_kwargs=dict(label=label),
        )
        b = CorrelationPlots.xcf(
            bkp2,
            region=region,
            r_factor=2,
            ax=ax,
            correlation_file=correlation_file2,
            mumin=mumin,
            mumax=mumax,
            plot_kwargs=dict(label=label2),
        )
        ax.grid(visible=True)
        ax.set_ylim(-1, 1)
        ax.legend()

        ax = axs[1]
        ax.plot(a[0], b[1] - a[1], c="k")
        ax.set_xlabel(axs[0].get_xlabel())
        ax.set_ylabel(f"{label2}-{label}")
        ax.grid(visible=True)
        ax.set_xlabel(None)

        ax = axs[2]
        ax.plot(a[0], b[2] / a[2], c="k")
        ax.set_xlabel(axs[0].get_xlabel())
        ax.set_ylabel(f"errors{label2}/{label}")
        ax.grid(visible=True)

        # axs[1].set_ylim(0.85, 1.15)
        axs[0].set_xlabel(None)

    @staticmethod
    def xcf(
        bookkeeper: Optional[Bookkeeper],
        region: str = "lya",
        absorber: str = "lya",
        mumin: float = 0,
        mumax: float = 1,
        abs_mu: bool = True,
        correlation_file: Path | str = "",
        ax: Optional[Axes] = None,
        r_factor: int = 2,
        plot_kwargs: Dict = dict(),
        just_return_values: bool = False,
        output_prefix: Optional[Path | str] = None,
        save_data: bool = False,
        save_plot: bool = False,
        save_dict: Dict = dict(),
        tracer: str = "qso",
    ) -> Tuple[np.ndarray, ...]:
        """
        Plotting correlation function in defined wedge.

        Arguments:
        ----------
        bookkeeper: bookkeeper object to use for collect data.
        region: region to use.
        absorber: absorber to use.
        correlation_file: Provide path to correlation file alternatively.
        mumin: wedge min value.
        mumax: wedge max value.
        ax: axis where to draw the plot, if None, it'll be created.
        r_factor: exponential factor to apply to distance r.
        plot_kwargs: extra kwargs to sent to the plotting function.
        output_prefix: Save the plot under this file structure (Default: None, plot not saved)
        save_data: Save the data into a npz file under the output_prefix file structure. (Default: False).
        save_plot: Save the plot into a png file. (Default: False).
        save_dict: Extra information to save in the npz file if save_data option is True. (Default: Empty dict)
        tracer: tracer to use.
        """
        if output_prefix is not None:
            output_prefix = Path(output_prefix)

        if isinstance(bookkeeper, Bookkeeper):
            correlation_file = bookkeeper.paths.exp_xcf_fname(absorber, region, tracer)

        with fitsio.FITS(correlation_file) as ffile:
            try:
                da = ffile["COR"]["DA"][:]
            except ValueError:
                da = ffile["COR"]["DA_BLIND"][:]
            co = ffile["COR"]["CO"][:]
            nb = ffile["COR"]["NB"][:]

            cor_header = ffile["COR"].read_header()
            wedge = Wedge(
                rp=(cor_header["RPMIN"], cor_header["RPMAX"], cor_header["NP"]),
                rt=(cor_header.get("RTMIN", 0), cor_header["RTMAX"], cor_header["NT"]),
                r=(cor_header.get("RTMIN", 0), cor_header["RTMAX"], cor_header["NT"]),
                mu=(mumin, mumax),
                abs_mu=abs_mu,
            )
        data_wedge = wedge(da, co)

        # Computing nb wedge manually
        axis_sum = wedge.weights.sum(axis=0)
        w = axis_sum > 0
        nb_weights = np.copy(wedge.weights)
        nb_weights[:, w] /= wedge.weights.sum(axis=0)[w]
        nb_wedge = data_wedge[0], nb_weights.dot(nb)

        r_coef = data_wedge[0] ** r_factor

        if just_return_values:
            return (
                data_wedge[0],
                r_coef * data_wedge[1],
                r_coef * np.sqrt(np.diag(data_wedge[2])),
                nb_wedge[1],
            )

        if (save_data or save_plot) and output_prefix is None:
            raise ValueError("Set output_prefix in order to save data.")
        if save_data:
            data_dict = {}
        if ax is None:
            fig, ax = plt.subplots()

        ax.errorbar(
            data_wedge[0],
            r_coef * data_wedge[1],
            yerr=r_coef * np.sqrt(np.diag(data_wedge[2])),
            **plot_kwargs,
        )
        ax.grid(visible=True)
        ax.set_xlabel(r"$r \, [\mathrm{Mpc \, h^{-1}}]$")
        ax.set_ylabel(
            r"$r^{0} \xi(r) \, [\mathrm{{Mpc \, h^{{-1}}  }}]$".format(r_factor)
        )
        ax.set_title("{0} < $\mu$ < {1}".format(mumin, mumax))

        if save_data:
            data_dict["r"] = data_wedge[0]
            data_dict["values"] = r_coef * data_wedge[1]
            data_dict["errors"] = r_coef * np.sqrt(np.diag(data_wedge[2]))
            data_dict["r_factor"] = r_factor
            data_dict["nb"] = nb_wedge[1]
            # data_dict['wedge_data'] = data_wedge

        plt.tight_layout()

        if output_prefix is not None:
            if save_plot:
                plt.savefig(
                    output_prefix.parent / (output_prefix.name + "-plot_xcf.png"),
                )
            if save_data:
                np.savez(
                    output_prefix.parent / (output_prefix.name + "-plot_xcf.npz"),
                    **{**save_dict, **data_dict},
                )
        elif save_data or save_plot:
            raise ValueError("Set output_prefix in order to save data.")

        return (
            data_wedge[0],
            r_coef * data_wedge[1],
            r_coef * np.sqrt(np.diag(data_wedge[2])),
            nb_wedge[1],
        )

    # @staticmethod
    # def multiple_cf(
    #     bookkeepers: List[Bookkeeper],
    #     region: str,
    #     labels: List[str] = None,
    #     region2: str = None,
    #     absorber: str = "lya",
    #     absorber2: str = None,
    #     correlation_file: Path | str = None,
    #     mumin: float = 0,
    #     mumax: float = 1,
    #     ax: Optional[Axes] = None,
    #     r_factor: int = 2,
    #     plot_kwargs: Dict = dict(),
    #     just_return_values: bool = False,
    #     output_prefix: Path | str = None,
    #     save_data: bool = False,
    #     save_plot: bool = False,
    #     save_dict: Dict = None,
    # ):
    #     """
    #     Plotting correlation function in defined wedge.

    #     Arguments:
    #     ----------
    #     bookkeepers: Bookkeepers object to use for collecting data.
    #     region: region to compute correlations from.
    #     labels: Labels to use for the different bookkeeper realizations.
    #     region2: region to compute correlations from if multiple regions have been used.
    #     mumin: wedge min value.
    #     mumax: wedge max value.
    #     ax: axis where to draw the plot, if None, it'll be created.
    #     r_factor: exponential factor to apply to distance r.
    #     output_prefix: Save the plot under this file structure (Default: None, plot not saved)
    #     save_data: Save the data into a npz file under the output_prefix file structure. (Default: False).
    #     save_plot: Save the plot into a png file. (Default: False).
    #     save_dict: Extra information to save in the npz file if save_data option is True. (Default: Empty dict)
    #     """
    #     if (save_data or save_plot) and output_prefix is None:
    #         raise ValueError("Set output_prefix in order to save data.")
    #     if save_data:
    #         data_dict = {}
    #     if ax is None:
    #         fig, ax = plt.subplots()

    #     for bookkeeper, label in zip(bookkeepers, labels):
    #         values = CorrelationPlots.plot_cf(
    #             bookkeeper=bookkeeper,
    #             region=region,
    #             region2=region2,
    #             mumin=mumin,
    #             mumax=mumax,
    #             plot_kwargs={**plot_kwargs, **dict(label=label)},
    #             ax=ax,
    #             r_factor=r_factor,
    #         )

    #         if save_data:
    #             data_dict[f"{label}_r"] = values[0]
    #             data_dict[f"{label}_values"] = values[1]
    #             data_dict[f"{label}_errors"] = values[2]
    #             data_dict[f"{label}_nb"] = values[3]

    #     ax.legend()
    #     plt.tight_layout()
    #     if save_plot:
    #         plt.savefig(
    #             output_prefix.parent / (output_prefix.name + "-multiple_cf.png"),
    #         )
    #     if save_data:
    #         np.savez(
    #             output_prefix.parent / (output_prefix.name + "-multiple_cf.npz"),
    #             **{**save_dict, **data_dict},
    #         )

    @staticmethod
    def cf_errorbarsize(
        bookkeeper: Bookkeeper,
        region: str,
        region2: Optional[str] = None,
        absorber: str = "lya",
        absorber2: Optional[str] = None,
        correlation_file: Optional[Path | str] = None,
        mumin: float = 0,
        mumax: float = 1,
        r_factor: int = 2,
        ax: Optional[Axes] = None,
        plot_kwargs: Dict = dict(),
        just_return_values: bool = False,
        output_prefix: Optional[Path | str] = None,
        save_data: bool = False,
        save_plot: bool = False,
        save_dict: Dict = dict(),
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Plotting correlation errorbar in defined wedge.

        Arguments:
        ----------
        bookkeeper: bookkeeper object to use for collect data.
        region: region to use.
        region2: if set use different second region.
        absorber: absorber to use.
        absorber2: if set use different second absorber.
        correlation_file: Provide path to correlation file alternatively.
        mumin: wedge min value.
        mumax: wedge max value.
        ax: axis where to draw the plot, if None, it'll be created.
        r_factor: exponential factor to apply to distance r.
        plot_kwargs: extra kwargs to sent to the plotting function.
        output_prefix: Save the plot under this file structure (Default: None, plot not saved)
        save_data: Save the data into a npz file under the output_prefix file structure. (Default: False).
        save_plot: Save the plot into a png file. (Default: False).
        save_dict: Extra information to save in the npz file if save_data option is True. (Default: Empty dict)
        """
        if output_prefix is not None:
            output_prefix = Path(output_prefix)

        with fitsio.FITS(
            bookkeeper.paths.exp_cf_fname(absorber, region, absorber2, region2)
        ) as ffile:
            try:
                da = ffile["COR"]["DA"][:]
            except ValueError:
                da = ffile["COR"]["DA_BLIND"][:]
            co = ffile["COR"]["CO"][:]
            nb = ffile["COR"]["NB"][:]

            cor_header = ffile["COR"].read_header()
            wedge = Wedge(
                rp=(cor_header["RPMIN"], cor_header["RPMAX"], cor_header["NP"]),
                rt=(cor_header.get("RTMIN", 0), cor_header["RTMAX"], cor_header["NT"]),
                r=(cor_header.get("RTMIN", 0), cor_header["RTMAX"], cor_header["NT"]),
                mu=(mumin, mumax),
            )
        data_wedge = wedge(da, co)

        # Computing nb wedge manually
        axis_sum = wedge.weights.sum(axis=0)
        w = axis_sum > 0
        nb_weights = np.copy(wedge.weights)
        nb_weights[:, w] /= wedge.weights.sum(axis=0)[w]
        nb_wedge = data_wedge[0], nb_weights.dot(nb)

        r_coef = data_wedge[0] ** r_factor

        if just_return_values:
            return data_wedge[0], r_coef * np.sqrt(np.diag(data_wedge[2])), nb_wedge[1]

        if (save_data or save_plot) and output_prefix is None:
            raise ValueError("Set output_prefix in order to save data.")
        if save_data:
            data_dict = {}
        if ax is None:
            fig, ax = plt.subplots()

        ax.plot(
            data_wedge[0],
            r_coef * np.sqrt(np.diag(data_wedge[2])),
            **plot_kwargs,
        )
        ax.grid(visible=True)
        ax.set_xlabel(r"$r \, [\mathrm{Mpc \, h^{-1}}]$")
        ax.set_ylabel(
            r"$r^{0} \sigma_\xi(r) \, [\mathrm{{Mpc \, h^{{-1}}  }}]$".format(r_factor)
        )
        ax.set_title("{0} < $\mu$ < {1}".format(mumin, mumax))
        plt.tight_layout()

        if save_data:
            data_dict["r"] = data_wedge[0]
            data_dict["values"] = r_coef * np.sqrt(np.diag(data_wedge[2]))
            data_dict["r_factor"] = r_factor
            data_dict["nb"] = nb_wedge[1]
            # data_dict['wedge_data'] = data_wedge

        if output_prefix is not None:
            if save_plot:
                plt.savefig(
                    output_prefix.parent / (output_prefix.name + "-plot_cf.png"),
                )
            if save_data:
                np.savez(
                    output_prefix.parent / (output_prefix.name + "-plot_cf.npz"),
                    **{**save_dict, **data_dict},
                )
        elif save_data or save_plot:
            raise ValueError("Set output_prefix in order to save data.")

        return data_wedge[0], r_coef * np.sqrt(np.diag(data_wedge[2])), nb_wedge[1]

    @staticmethod
    def xcf_errorbarsize(
        bookkeeper: Bookkeeper,
        region: str,
        absorber: str = "lya",
        correlation_file: Optional[Path | str] = None,
        mumin: float = 0,
        mumax: float = 1,
        abs_mu: bool = True,
        r_factor: int = 2,
        ax: Optional[Axes] = None,
        plot_kwargs: Dict = dict(),
        just_return_values: bool = False,
        output_prefix: Optional[Path | str] = None,
        save_data: bool = False,
        save_plot: bool = False,
        save_dict: Dict = dict(),
        tracer: str = "qso",
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Plotting cross-correlation errorbars in defined wedge.

        Arguments:
        ----------
        bookkeeper: bookkeeper object to use for collect data.
        region: region to use.
        absorber: absorber to use.
        correlation_file: Provide path to correlation file alternatively.
        mumin: wedge min value.
        mumax: wedge max value.
        ax: axis where to draw the plot, if None, it'll be created.
        r_factor: exponential factor to apply to distance r.
        plot_kwargs: extra kwargs to sent to the plotting function.
        output_prefix: Save the plot under this file structure (Default: None, plot not saved)
        save_data: Save the data into a npz file under the output_prefix file structure. (Default: False).
        save_plot: Save the plot into a png file. (Default: False).
        save_dict: Extra information to save in the npz file if save_data option is True. (Default: Empty dict)
        tracer: tracer to use.
        """
        if output_prefix is not None:
            output_prefix = Path(output_prefix)

        with fitsio.FITS(
            bookkeeper.paths.exp_xcf_fname(absorber, region, tracer)
        ) as ffile:
            try:
                da = ffile["COR"]["DA"][:]
            except ValueError:
                da = ffile["COR"]["DA_BLIND"][:]
            co = ffile["COR"]["CO"][:]
            nb = ffile["COR"]["NB"][:]

            cor_header = ffile["COR"].read_header()
            wedge = Wedge(
                rp=(cor_header["RPMIN"], cor_header["RPMAX"], cor_header["NP"]),
                rt=(cor_header.get("RTMIN", 0), cor_header["RTMAX"], cor_header["NT"]),
                r=(cor_header.get("RTMIN", 0), cor_header["RTMAX"], cor_header["NT"]),
                mu=(mumin, mumax),
                abs_mu=abs_mu,
            )
        data_wedge = wedge(da, co)

        # Computing nb wedge manually
        axis_sum = wedge.weights.sum(axis=0)
        w = axis_sum > 0
        nb_weights = np.copy(wedge.weights)
        nb_weights[:, w] /= wedge.weights.sum(axis=0)[w]
        nb_wedge = data_wedge[0], nb_weights.dot(nb)

        r_coef = data_wedge[0] ** r_factor

        if just_return_values:
            return data_wedge[0], r_coef * np.sqrt(np.diag(data_wedge[2])), nb_wedge[1]

        if (save_data or save_plot) and output_prefix is None:
            raise ValueError("Set output_prefix in order to save data.")
        if save_data:
            data_dict = {}
        if ax is None:
            fig, ax = plt.subplots()

        ax.plot(
            data_wedge[0],
            r_coef * np.sqrt(np.diag(data_wedge[2])),
            **plot_kwargs,
        )
        ax.grid(visible=True)
        ax.set_xlabel(r"$r \, [\mathrm{Mpc \, h^{-1}}]$")
        ax.set_ylabel(
            r"$r^{0} \sigma_\xi(r) \, [\mathrm{{Mpc \, h^{{-1}}  }}]$".format(r_factor)
        )
        ax.set_title("{0} < $\mu$ < {1}".format(mumin, mumax))
        plt.tight_layout()

        if save_data:
            data_dict["r"] = data_wedge[0]
            data_dict["values"] = r_coef * np.sqrt(np.diag(data_wedge[2]))
            data_dict["r_factor"] = r_factor
            data_dict["nb"] = nb_wedge[1]
            # data_dict['wedge_data'] = data_wedge

        if output_prefix is not None:
            if save_plot:
                plt.savefig(
                    output_prefix.parent / (output_prefix.name + "-plot_cf.png"),
                )
            if save_data:
                np.savez(
                    output_prefix.parent / (output_prefix.name + "-plot_cf.npz"),
                    **{**save_dict, **data_dict},
                )
        elif save_data or save_plot:
            raise ValueError("Set output_prefix in order to save data.")

        return data_wedge[0], r_coef * np.sqrt(np.diag(data_wedge[2])), nb_wedge[1]

    # @staticmethod
    # def multiple_cf_errorbarsize(
    #     bookkeepers: List[Bookkeeper],
    #     region: str,
    #     labels: List[str] = None,
    #     region2: str = None,
    #     absorber: str = "lya",
    #     absorber2: str = None,
    #     correlation_file: Path | str = None,
    #     mumin: float = 0,
    #     mumax: float = 1,
    #     ax: Optional[Axes] = None,
    #     r_factor: int = 2,
    #     plot_kwargs: Dict = dict(),
    #     just_return_values: bool = False,
    #     output_prefix: Path | str = None,
    #     save_data: bool = False,
    #     save_plot: bool = False,
    #     save_dict: Dict = None,
    # ):
    #     """
    #     Plotting correlation function errors in defined wedge.

    #     Arguments:
    #     ----------

    #     bookkeepers: Bookkeepers object to use for collecting data.
    #     region: region to compute correlations from.
    #     labels: Labels to use for the different bookkeeper realizations.
    #     region2: region to compute correlations from if multiple regions have been used.
    #     mumin: wedge min value.
    #     mumax: wedge max value.
    #     ax: axis where to draw the plot, if None, it'll be created.
    #     r_factor: exponential factor to apply to distance r.
    #     output_prefix: Save the plot under this file structure (Default: None, plot not saved)
    #     save_data: Save the data into a npz file under the output_prefix file structure. (Default: False).
    #     save_plot: Save the plot into a png file. (Default: False).
    #     save_dict: Extra information to save in the npz file if save_data option is True. (Default: Empty dict)
    #     """
    #     if (save_data or save_plot) and output_prefix is None:
    #         raise ValueError("Set output_prefix in order to save data.")
    #     if save_data:
    #         data_dict = {}
    #     if ax is None:
    #         fig, ax = plt.subplots()

    #     for bookkeeper, label in zip(bookkeepers, labels):
    #         values = CorrelationPlots.plot_cf_errorbarsize(
    #             bookkeeper=bookkeeper,
    #             region=region,
    #             region2=region2,
    #             mumin=mumin,
    #             mumax=mumax,
    #             ax=ax,
    #             r_factor=r_factor,
    #             plot_kwargs={**plot_kwargs, **dict(label=label)},
    #         )

    #         if save_data:
    #             data_dict[f"{label}_r"] = values[0]
    #             data_dict[f"{label}_errors"] = values[1]

    #     ax.legend()
    #     plt.tight_layout()
    #     if save_plot:
    #         plt.savefig(
    #             output_prefix.parent / (output_prefix.name + "-multiple_cf.png"),
    #         )
    #     if save_data:
    #         np.savez(
    #             output_prefix.parent / (output_prefix.name + "-multiple_cf.npz"),
    #             **{**save_dict, **data_dict},
    #         )

    @staticmethod
    def cf_map(
        bookkeeper: Bookkeeper,
        region: str = "lya",
        region2: Optional[str] = None,
        absorber: str = "lya",
        absorber2: Optional[str] = None,
        correlation_file: Optional[Path | str] = None,
        r_factor: int = 2,
        vmin: float = -0.04,
        vmax: float = 0.04,
        fig: Optional[Figure] = None,
        ax: Optional[Axes] = None,
        plot_kwargs: Dict = dict(),
        just_return_values: bool = False,
        output_prefix: Optional[Path | str] = None,
        save_data: bool = False,
        save_plot: bool = False,
        save_dict: Dict = dict(),
    ) -> Tuple[Tuple[float, ...], np.ndarray, np.ndarray, np.ndarray]:
        """
        Plotting correlation heatmap.

        Arguments:
        ----------
        bookkeeper: bookkeeper object to use for collect data.
        region: region to use.
        region2: if set use different second region.
        absorber: absorber to use.
        absorber2: if set use different second absorber.
        correlation_file: Provide path to correlation file alternatively.
        r_factor: exponential factor to apply to distance r.
        vmin: min third axis for the colormap.
        vmax: max third axis for the colormap.
        fig: matplotlib figure to make the plot in.
        ax: axis where to draw the plot, if None, it'll be created.
        plot_kwargs: extra kwargs to sent to the plotting function.
        output_prefix: Save the plot under this file structure (Default: None, plot not saved)
        save_data: Save the data into a npz file under the output_prefix file structure. (Default: False).
        save_plot: Save the plot into a png file. (Default: False).
        save_dict: Extra information to save in the npz file if save_data option is True. (Default: Empty dict)
        """
        if output_prefix is not None:
            output_prefix = Path(output_prefix)

        with fitsio.FITS(
            bookkeeper.paths.exp_cf_fname(absorber, region, absorber2, region2)
        ) as ffile:
            try:
                da = ffile["COR"]["DA"][:]
            except ValueError:
                da = ffile["COR"]["DA_BLIND"][:]
            co = ffile["COR"]["CO"][:]
            nb = ffile["COR"]["NB"][:]
            rp = ffile["COR"]["RP"][:]
            rt = ffile["COR"]["RT"][:]

            cor_header = ffile["COR"].read_header()

        extent: tuple[float, ...]
        extent = (
            cor_header.get("RTMIN", 0),
            cor_header["RTMAX"],
            cor_header["RPMIN"],
            cor_header["RPMAX"],
        )
        r = np.sqrt(rp**2 + rt**2)
        nrp, nrt = cor_header["NP"], cor_header["NT"]
        r = r.reshape(nrp, nrt)
        mat = da.reshape(nrp, nrt) * r**r_factor
        errmat = np.sqrt(np.diag(co)).reshape(nrp, nrt) * r**r_factor
        nmat = nb.reshape(nrp, nrt)

        if just_return_values:
            return extent, mat, errmat, nmat

        if (save_data or save_plot) and output_prefix is None:
            raise ValueError("Set output_prefix in order to save data.")
        if save_data:
            data_dict = dict(extent=extent, mat=mat, errmat=errmat, nmat=nmat)
        if ax is None:
            fig, ax = plt.subplots()
        elif fig is None:
            raise ValueError("ax and fiug should be provided at the same time")

        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)

        im = ax.imshow(
            mat,
            origin="lower",
            extent=extent,
            interpolation="nearest",
            cmap="seismic",
            **{
                **dict(vmin=vmin, vmax=vmax),
                **plot_kwargs,
            },
        )

        fig.colorbar(im, cax=cax, orientation="vertical")
        ax.set_xlabel(r"$r_{\perp} \, [h^{-1} \, \rm{Mpc}]$")
        ax.set_ylabel(r"$r_{\parallel} \, [h^{-1} \, \rm{Mpc}]$")
        cax.yaxis.set_label_position("right")
        cax.set_ylabel(r"$r\xi(r_{\parallel,r_{\perp}})$")

        if output_prefix is not None:
            if save_plot:
                output_prefix = Path(output_prefix)
                plt.tight_layout()
                plt.savefig(
                    str(output_prefix) + ".png",
                    dpi=300,
                )

            if save_data:
                np.savez(
                    str(output_prefix) + ".npz",
                    **{**save_dict, **data_dict},
                )
        elif save_data or save_plot:
            raise ValueError("Set output_prefix in order to save data.")

        return extent, mat, errmat, nmat

    @staticmethod
    def xcf_map(
        bookkeeper: Bookkeeper,
        region: str = "lya",
        absorber: str = "lya",
        absorber2: Optional[str] = None,
        correlation_file: Optional[Path | str] = None,
        r_factor: int = 2,
        vmin: float = -0.4,
        vmax: float = 0.4,
        fig: Optional[Figure] = None,
        ax: Optional[Axes] = None,
        plot_kwargs: Dict = dict(),
        just_return_values: bool = False,
        output_prefix: Optional[Path | str] = None,
        save_data: bool = False,
        save_plot: bool = False,
        save_dict: Dict = dict(),
        tracer: str = "qso",
    ) -> Tuple[Tuple[float, ...], np.ndarray, np.ndarray, np.ndarray]:
        """
        Plotting correlation heatmap.

        Arguments:
        ----------
        bookkeeper: Bookkeeper object to use for collect data.
        region: region to compute correlations from.
        r_factor: exponential factor to apply to distance r.
        ax: axis where to draw the plot, if None, it'll be created.
        output_prefix: Save the plot under this file structure (Default: None, plot not saved)
        save_data: Save the data into a npz file under the output_prefix file structure. (Default: False).
        save_plot: Save the plot into a png file. (Default: False).
        save_dict: Extra information to save in the npz file if save_data option is True. (Default: Empty dict)
        tracer: Tracer to use.
        """
        if output_prefix is not None:
            output_prefix = Path(output_prefix)

        with fitsio.FITS(
            bookkeeper.paths.exp_xcf_fname(absorber, region, tracer)
        ) as ffile:
            try:
                da = ffile["COR"]["DA"][:]
            except ValueError:
                da = ffile["COR"]["DA_BLIND"][:]
            co = ffile["COR"]["CO"][:]
            nb = ffile["COR"]["NB"][:]
            rp = ffile["COR"]["RP"][:]
            rt = ffile["COR"]["RT"][:]

            cor_header = ffile["COR"].read_header()

        extent: tuple[float, ...]
        extent = (
            cor_header.get("RTMIN", 0),
            cor_header["RTMAX"],
            cor_header["RPMIN"],
            cor_header["RPMAX"],
        )
        r = np.sqrt(rp**2 + rt**2)
        nrp, nrt = cor_header["NP"], cor_header["NT"]
        r = r.reshape(nrp, nrt)
        mat = da.reshape(nrp, nrt) * r**r_factor
        errmat = np.sqrt(np.diag(co)).reshape(nrp, nrt) * r**r_factor
        nmat = nb.reshape(nrp, nrt)

        if just_return_values:
            return extent, mat, errmat, nmat

        if (save_data or save_plot) and output_prefix is None:
            raise ValueError("Set output_prefix in order to save data.")
        if save_data:
            data_dict = dict(extent=extent, mat=mat, errmat=errmat, nmat=nmat)
        if ax is None:
            fig, ax = plt.subplots()
        elif fig is None:
            raise ValueError("ax and fiug should be provided at the same time")

        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)

        im = ax.imshow(
            mat,
            origin="lower",
            extent=extent,
            interpolation="nearest",
            cmap="seismic",
            **{
                **dict(vmin=vmin, vmax=vmax),
                **plot_kwargs,
            },
        )

        fig.colorbar(im, cax=cax, orientation="vertical")
        ax.set_xlabel(r"$r_{\perp} \, [h^{-1} \, \rm{Mpc}]$")
        ax.set_ylabel(r"$r_{\parallel} \, [h^{-1} \, \rm{Mpc}]$")
        cax.yaxis.set_label_position("right")
        cax.set_ylabel(r"$r\xi(r_{\parallel,r_{\perp}})$")

        if output_prefix is not None:
            if save_plot:
                output_prefix = Path(output_prefix)
                plt.tight_layout()
                plt.savefig(
                    str(output_prefix) + ".png",
                    dpi=300,
                )

            if save_data:
                np.savez(
                    str(output_prefix) + ".npz",
                    **{**save_dict, **data_dict},
                )
        elif save_data or save_plot:
            raise ValueError("Set output_prefix in order to save data.")

        return extent, mat, errmat, nmat



"""Utils."""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

import fitsio
import healpy as hp
import numpy as np

from picca_bookkeeper.constants import forest_regions

if TYPE_CHECKING:
    from typing import Dict, List, Optional, Tuple

    from picca_bookkeeper.hints import wave_grid, wave_grid_int, wave_grid_rf


def find_bins(original_array: wave_grid, grid_array: wave_grid) -> wave_grid_int:
    idx = np.searchsorted(grid_array, original_array)
    np.clip(idx, 0, len(grid_array) - 1, out=idx)

    prev_index_closer = (grid_array[idx - 1] - original_array) ** 2 <= (
        grid_array[idx] - original_array
    ) ** 2
    return idx - prev_index_closer


def find_qso_pixel(los_id: int, catalog: str | Path) -> Tuple[int, str]:
    """Find healpix pixel where given quasar is located

    Arguments
    ---------
    los_id: Line of sight id of quasar.
    catalog: Path to quasar catalog.
    """
    with fitsio.FITS(catalog) as hdul:
        indx = np.where(hdul["ZCATALOG"]["TARGETID"][:] == los_id)[0][0]

        ra = hdul["ZCATALOG"]["TARGET_RA"][indx]
        dec = hdul["ZCATALOG"]["TARGET_DEC"][indx]

        return (
            hp.pixelfunc.ang2pix(64, ra, dec, lonlat=True, nest=True),
            hdul["ZCATALOG"]["SURVEY"][indx],
        )


def compute_zeff(
    export_files: List[Path],
    rmins: List[float] | float = 0,
    rmaxs: List[float] | float = 1000,
) -> float:
    """Compute zeff from a set of export files"""
    if not isinstance(rmins, list) or (isinstance(rmins, list) and len(rmins) == 1):
        rmins = [rmins for file in export_files]  # type: ignore

    if not isinstance(rmaxs, list) or (isinstance(rmaxs, list) and len(rmaxs) == 1):
        rmaxs = [rmaxs for file in export_files]  # type: ignore

    zeff_list = []
    weights = []
    for export_file, rmin, rmax in zip(export_files, rmins, rmaxs):
        with fitsio.FITS(export_file) as hdul:
            r_arr = np.sqrt(hdul[1].read()["RP"] ** 2 + hdul[1].read()["RT"] ** 2)
            cells = (r_arr > rmin) * (r_arr < rmax)

            inverse_variance = 1 / np.diag(hdul[1].read()["CO"])
            zeff = np.average(
                hdul[1].read()["Z"][cells], weights=inverse_variance[cells]
            )
            weight = np.sum(inverse_variance[cells])

        zeff_list.append(zeff)
        weights.append(weight)

    return np.average(zeff_list, weights=weights)


def get_spectra_from_los_id(
    los_id: int,
    catalog: str | Path,
    input_healpix: Optional[str | Path] = None,
    lambda_grid: wave_grid = np.arange(3600, 9824, 0.8),
) -> Tuple[wave_grid, wave_grid, wave_grid]:
    """Get quasar spectra given los_id

    Arguments
    ---------
    los_id: Line of sight id of quasar.
    catalog: Path to quasar catalog.
    lambda_grid: Grid to show spectra for (np.array).
    """
    healpix, survey = find_qso_pixel(los_id, catalog)
    healpix_str = str(healpix)

    if input_healpix is None:
        input_healpix = Path("/global/cfs/cdirs/desi/science/lya/fugu_healpix/healpix")
    else:
        input_healpix = Path(input_healpix)

    coadd_file = (
        input_healpix
        / survey
        / "dark"
        / healpix_str[:-2]
        / healpix_str
        / f"coadd-{survey}-dark-{healpix_str}.fits"
    )

    assert coadd_file.is_file()

    with fitsio.FITS(coadd_file) as hdul:
        assert los_id in hdul["FIBERMAP"]["TARGETID"]

        idx = np.where(hdul["FIBERMAP"]["TARGETID"][:] == los_id)[0][0]

        flux_data = dict(
            B_WAVE=hdul["B_WAVELENGTH"].read(),
            R_WAVE=hdul["R_WAVELENGTH"].read(),
            Z_WAVE=hdul["Z_WAVELENGTH"].read(),
            B_FLUX=hdul["B_FLUX"].read()[idx],
            R_FLUX=hdul["R_FLUX"].read()[idx],
            Z_FLUX=hdul["Z_FLUX"].read()[idx],
            B_MASK=hdul["B_MASK"].read()[idx],
            R_MASK=hdul["R_MASK"].read()[idx],
            Z_MASK=hdul["Z_MASK"].read()[idx],
            B_IVAR=hdul["B_IVAR"].read()[idx] * (hdul["B_MASK"].read()[idx] == 0),
            R_IVAR=hdul["R_IVAR"].read()[idx] * (hdul["R_MASK"].read()[idx] == 0),
            Z_IVAR=hdul["Z_IVAR"].read()[idx] * (hdul["Z_MASK"].read()[idx] == 0),
        )

    return coadd_flux_data(flux_data, lambda_grid)


def coadd_flux_data(
    flux_data: Dict, lambda_grid: wave_grid
) -> Tuple[wave_grid, wave_grid, wave_grid]:
    """Coadd flux data for different arms

    Arguments
    ---------
    flux_data: Dictionary containing flux data in keys 'B', 'R', 'Z'.
    lambda_grid: Grid to show spectra for (np.array).
    """
    flux = np.zeros_like(lambda_grid)
    ivar = np.zeros_like(lambda_grid)

    for color in "B", "R", "Z":
        bins = find_bins(flux_data[f"{color}_WAVE"], lambda_grid)

        flux += np.bincount(
            bins,
            weights=flux_data[f"{color}_IVAR"] * flux_data[f"{color}_FLUX"],
            minlength=flux.size,
        )
        ivar += np.bincount(
            bins, weights=flux_data[f"{color}_IVAR"], minlength=ivar.size
        )

    w = ivar > 0
    flux[w] /= ivar[w]

    return lambda_grid, flux, ivar


def Ulambda(lambda_rest: wave_grid, region: str = "lya") -> wave_grid_rf:
    """Helper function to buid quasar continuum from aq bq parameters"""
    lambda_min = forest_regions[region]["lambda-rest-min"]
    lambda_max = forest_regions[region]["lambda-rest-max"]

    return (np.log10(lambda_rest) - np.log10(lambda_min)) / (
        np.log10(lambda_max) - np.log10(lambda_min)
    )


def compute_cont(
    los_id: int, attrs_file: str | Path, region: str = "lya"
) -> Tuple[wave_grid_rf, wave_grid_rf]:
    """Compute quasar continuum using attributes file (will read aq, bq and mean cont)

    Arguments:
    ----------
    los_id: Line of sight id of quasar.
    attrs_file: Path to delta attributes file.
    region: Region where to compute the continuum.
    """
    with fitsio.FITS(attrs_file) as hdul:
        lambda_rest = 10 ** hdul["CONT"]["loglam_rest"][:]
        mean_cont = hdul["CONT"]["mean_cont"][:]

        idx = np.where(hdul["FIT_METADATA"]["LOS_ID"][:] == los_id)[0][0]
        aq = hdul["FIT_METADATA"]["ZERO_POINT"][idx]
        bq = hdul["FIT_METADATA"]["SLOPE"][idx]

        cont = mean_cont * (aq + bq * (Ulambda(lambda_rest, region=region)))

    return lambda_rest, cont



from __future__ import annotations

import itertools
import os
import warnings
from configparser import ConfigParser
from multiprocessing import Pool
from pathlib import Path
from typing import TYPE_CHECKING

import fitsio
import matplotlib.pyplot as plt
import numpy as np
from scipy.interpolate import interp1d

from picca_bookkeeper.bookkeeper import Bookkeeper
from picca_bookkeeper.utils import compute_cont, get_spectra_from_los_id

if TYPE_CHECKING:
    from typing import Any, Dict, List, Optional, Tuple

    from picca_bookkeeper.hints import Axes, wave_grid, wave_grid_int, wave_grid_rf


class ReadDeltas:
    def __init__(
        self,
        bookkeeper: Bookkeeper | Path | str,
        region: Optional[str] = None,
        calib_step: Optional[int] = None,
        blind: bool = False,
        label: Optional[str] = None,
    ):
        self.region = region
        self.calib_step = calib_step

        if isinstance(bookkeeper, Bookkeeper):
            self.bookkeeper = bookkeeper
        else:
            self.bookkeeper = Bookkeeper(bookkeeper)

        self.blind = blind
        self.set_var_lss_mod()
        self.define_lambda_arrays()
        self.define_interp_quantities()
        self.label = label

    def __str__(self) -> str:
        return str(self.label)

    @property
    def config_file_in(self) -> Path:
        if self.calib_step is not None:
            return (
                self.bookkeeper.paths.run_path
                / "configs"
                / f"delta_extraction_{self.region}_calib_step_{self.calib_step}.ini"
            )
        else:
            return (
                self.bookkeeper.paths.run_path
                / "configs"
                / f"delta_extraction_{self.region}.ini"
            )

    @property
    def config_file(self) -> Path:
        return self.deltas_path.parent / ".config.ini"

    @property
    def deltas_path(self) -> Path:
        return self.bookkeeper.paths.deltas_path(
            region=self.region, calib_step=self.calib_step
        )

    @property
    def attributes_file(self) -> Path:
        return self.bookkeeper.paths.delta_attributes_file(
            region=self.region, calib_step=self.calib_step
        )

    def set_var_lss_mod(self) -> None:
        config = ConfigParser()
        config.read(self.config_file)
        self.var_lss_mod = config["expected flux"].getfloat("var lss mod", 1)

    def define_lambda_arrays(self) -> None:
        config = ConfigParser()
        config.read(self.config_file)

        if config["data"]["wave solution"] == "lin":
            self.lambda_grid = np.arange(
                float(config["data"]["lambda min"]),
                float(config["data"]["lambda max"])
                + float(config["data"]["delta lambda"]),
                float(config["data"]["delta lambda"]),
            )

            self.lambda_rf_grid = np.arange(
                float(config["data"]["lambda min rest frame"]),
                float(config["data"]["lambda max rest frame"])
                + float(config["data"]["delta lambda"]),
                float(config["data"]["delta lambda"]),
            )

        else:
            self.lambda_grid = np.arange(
                np.log10(float(config["data"]["lambda min"])),
                np.log10(float(config["data"]["lambda max"]))
                + float(config["data"]["delta log lambda"]),
                float(config["data"]["delta log lambda"]),
            )
            self.lambda_grid = 10**self.lambda_grid

            self.lambda_rf_grid = np.arange(
                np.log10(float(config["data"]["lambda min rest frame"])),
                np.log10(float(config["data"]["lambda max rest frame"]))
                + float(config["data"]["delta log lambda"]),
                float(config["data"]["delta log lambda"]),
            )
            self.lambda_rf_grid = 10**self.lambda_grid

        self.lambda_edges: np.ndarray = np.concatenate(
            (  # type: ignore
                ((1.5 * self.lambda_grid[0] - 0.5 * self.lambda_grid[1]),),
                0.5 * (self.lambda_grid[1:] + self.lambda_grid[:-1]),
                ((1.5 * self.lambda_grid[-1] - 0.5 * self.lambda_grid[-2]),),
            )
        )

        self.lambda_rf_edges: np.ndarray = np.concatenate(
            (  # type: ignore
                ((1.5 * self.lambda_rf_grid[0] - 0.5 * self.lambda_rf_grid[1]),),
                0.5 * (self.lambda_rf_grid[1:] + self.lambda_rf_grid[:-1]),
                ((1.5 * self.lambda_rf_grid[-1] - 0.5 * self.lambda_rf_grid[-2]),),
            )
        )

    def define_interp_quantities(self) -> None:
        with fitsio.FITS(self.attributes_file) as attrs_fits:
            if "ETA" in [
                name.upper() for name in attrs_fits["VAR_FUNC"].get_colnames()
            ]:
                self.eta_interp = interp1d(
                    10 ** attrs_fits["VAR_FUNC"]["loglam"][:],
                    attrs_fits["VAR_FUNC"]["eta"][:],
                    fill_value="extrapolate",
                )
                self.eta = self.eta_interp(self.lambda_grid)
            else:
                self.eta = np.ones_like(self.lambda_grid)

            if "FUDGE" in [
                name.upper() for name in attrs_fits["VAR_FUNC"].get_colnames()
            ]:
                self.fudge_interp = interp1d(
                    10 ** attrs_fits["VAR_FUNC"]["loglam"][:],
                    attrs_fits["VAR_FUNC"]["fudge"][:],
                    fill_value="extrapolate",
                )
                self.fudge = self.fudge_interp(self.lambda_grid)
            else:
                self.fudge = 0 * np.ones_like(self.lambda_grid)

            self.var_lss_interp = interp1d(
                10 ** attrs_fits["VAR_FUNC"]["loglam"][:],
                attrs_fits["VAR_FUNC"]["var_lss"][:],
                fill_value="extrapolate",
            )
            self.var_lss = self.var_lss_interp(self.lambda_grid)

            self.mean_cont_interp = interp1d(
                10 ** attrs_fits["CONT"]["loglam_rest"][:],
                attrs_fits["CONT"]["mean_cont"][:],
                fill_value="extrapolate",
            )

            self.mean_cont_weight_interp = interp1d(
                10 ** attrs_fits["CONT"]["loglam_rest"][:],
                attrs_fits["CONT"]["weight"][:],
                fill_value="extrapolate",
            )

            self.mean_cont = self.mean_cont_interp(self.lambda_rf_grid)
            self.mean_cont_weight = self.mean_cont_weight_interp(self.lambda_rf_grid)

            if "STACK_DELTAS" in attrs_fits:
                self.delta_stack_interp = interp1d(
                    10 ** attrs_fits["STACK_DELTAS"]["loglam"][:],
                    attrs_fits["STACK_DELTAS"]["stack"][:],
                    fill_value="extrapolate",
                )

                self.delta_stack_weight_interp = interp1d(
                    10 ** attrs_fits["STACK_DELTAS"]["loglam"][:],
                    attrs_fits["STACK_DELTAS"]["weight"][:],
                    fill_value="extrapolate",
                )

                self.delta_stack = self.delta_stack_interp(self.lambda_grid)
                self.delta_stack_weight = self.delta_stack_weight_interp(
                    self.lambda_grid
                )

    def get_statistics(
        self, downsampling: float = 1, random_seed: int = 0, read_flux: bool = False
    ) -> None:
        delta_files = list(self.deltas_path.glob("delta*fits*"))

        pool = Pool(processes=5)
        self.statistics = pool.starmap(
            self.get_file_statistics,
            zip(
                delta_files,
                itertools.repeat(downsampling),
                itertools.repeat(random_seed),
                itertools.repeat(read_flux),
            ),
        )
        pool.close()

        self.deltas_arr = np.vstack(np.asarray(self.statistics, dtype="object").T[0])
        self.var_pipe_arr = np.vstack(np.asarray(self.statistics, dtype="object").T[1])
        self.weights_arr = np.vstack(np.asarray(self.statistics, dtype="object").T[2])
        self.lambda_rf_arr = np.vstack(np.asarray(self.statistics, dtype="object").T[3])
        self.id_arr = np.concatenate(np.asarray(self.statistics, dtype="object").T[4])
        self.cont_arr = np.vstack(np.asarray(self.statistics, dtype="object").T[5])
        self.flux_arr = np.vstack(np.asarray(self.statistics, dtype="object").T[6])
        self.flux_ivar_arr = np.vstack(np.asarray(self.statistics, dtype="object").T[7])
        self.z_arr = np.concatenate(np.asarray(self.statistics, dtype="object").T[8])
        self.meansnr_arr = np.concatenate(
            np.asarray(self.statistics, dtype="object").T[9]
        )
        self.deltas2_arr = self.deltas_arr**2

    def get_file_statistics(
        self,
        deltas_filename: Path,
        downsampling: float = 1,
        random_seed: int = 0,
        read_flux: bool = False,
    ) -> Tuple[np.ndarray, ...]:
        delta_field = "DELTA" if not self.blind else "DELTA_BLIND"
        with fitsio.FITS(deltas_filename) as delta_fits:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                VAR = 1 / delta_fits["WEIGHT"].read()
                var_pipe = (
                    VAR
                    - self.var_lss * self.var_lss_mod
                    + np.sqrt(
                        (self.var_lss * self.var_lss_mod - VAR) ** 2
                        - 4 * self.eta * self.fudge
                    )
                ) / (2 * self.eta)
                # var_pipe = (VAR - self.var_lss) / self.eta

            lambda_rf = np.asarray(
                [self.lambda_grid / (1 + x) for x in delta_fits["METADATA"]["Z"][:]]
            )

            if downsampling != 1:
                np.random.seed(random_seed)
                msk = np.random.choice(
                    [True, False],
                    delta_fits["METADATA"].read_header()["NAXIS2"],
                    p=[downsampling, 1 - downsampling],
                )
            else:
                msk = np.ones(
                    delta_fits["METADATA"].read_header()["NAXIS2"], dtype=bool
                )

            if not read_flux:
                return (
                    delta_fits[delta_field].read()[msk],
                    var_pipe[msk],
                    delta_fits["WEIGHT"].read()[msk],
                    lambda_rf[msk],
                    delta_fits["METADATA"]["LOS_ID"][:][msk],
                    delta_fits["CONT"].read()[msk],
                    np.zeros_like(lambda_rf[msk]),
                    np.zeros_like(lambda_rf[msk]),
                    delta_fits["METADATA"]["Z"][:][msk],
                    delta_fits["METADATA"]["MEANSNR"][:][msk],
                )
            else:
                healpix = deltas_filename.name[6:]
                with fitsio.FITS(
                    deltas_filename.parent.parent / "Flux" / f"flux-info-{healpix}"
                ) as flux_fits:
                    return (
                        delta_fits[delta_field].read()[msk],
                        var_pipe[msk],
                        delta_fits["WEIGHT"].read()[msk],
                        lambda_rf[msk],
                        delta_fits["METADATA"]["LOS_ID"][:][msk],
                        delta_fits["CONT"].read()[msk],
                        flux_fits["FLUX"].read()[msk],
                        flux_fits["FLUX_IVAR"].read()[msk],
                        delta_fits["METADATA"]["Z"][:][msk],
                        delta_fits["METADATA"]["MEANSNR"][:][msk],
                    )


class ReadDeltasNoBookkeeper(ReadDeltas):
    def __init__(
        self,
        deltas_path: Path | str,
        log_path: Path | str,
        config_file: Path | str,
        region: str,
        config_file_in: Optional[Path | str] = None,
        blind: bool = False,
        label: Optional[str] = None,
    ):
        self._config_file = Path(config_file)
        self._config_file_in = (
            Path(config_file_in) if config_file_in is not None else Path(config_file)
        )
        self._attributes_file = Path(log_path) / "delta_attributes.fits.gz"
        self._deltas_path = Path(deltas_path)

        super().__init__(
            bookkeeper=Path(deltas_path), region=region, blind=blind, label=label
        )

    @property
    def config_file_in(self) -> Path:
        return self._config_file_in

    @property
    def config_file(self) -> Path:
        return self._config_file

    @property
    def deltas_path(self) -> Path:
        return self._deltas_path

    @property
    def attributes_file(self) -> Path:
        return self._attributes_file


class Plots:
    @staticmethod
    def eta(
        bookkeeper: Optional[Bookkeeper] = None,
        region: Optional[str] = None,
        calib_step: Optional[int] = None,
        attributes_file: Optional[Path | str] = None,
        ax: Optional[Axes] = None,
        plot_kwargs: Dict = dict(),
    ) -> Tuple[wave_grid, wave_grid]:
        if ax is None:
            fig, ax = plt.subplots()

        if isinstance(bookkeeper, Bookkeeper):
            attributes_file = bookkeeper.paths.delta_attributes_file(
                region=region, calib_step=calib_step
            )

        with fitsio.FITS(attributes_file) as hdul:
            if "VAR_FUNC" in hdul:
                card = "VAR_FUNC"
            else:
                card = "WEIGHT"

            wave = 10 ** hdul[card]["LOGLAM"].read()
            eta = hdul[card]["ETA"].read()
            ax.plot(
                wave,
                eta,
                **plot_kwargs,
            )

        ax.set_xlabel(r"$\lambda \, [\AA]$")
        ax.set_ylabel(r"$\eta$")

        return wave, eta

    @staticmethod
    def var_lss(
        bookkeeper: Optional[Bookkeeper] = None,
        region: Optional[str] = None,
        calib_step: Optional[int] = None,
        attributes_file: Optional[Path | str] = None,
        ax: Optional[Axes] = None,
        plot_kwargs: Dict = dict(),
    ) -> Tuple[wave_grid, wave_grid]:
        if ax is None:
            fig, ax = plt.subplots()

        if isinstance(bookkeeper, Bookkeeper):
            attributes_file = bookkeeper.paths.delta_attributes_file(
                region=region, calib_step=calib_step
            )

        with fitsio.FITS(attributes_file) as hdul:
            if "STATS" in hdul:
                wave = hdul["STATS"]["LAMBDA"].read()
                var_lss = hdul["STATS"]["VAR"].read()
            else:
                if "VAR_FUNC" in hdul:
                    card = "VAR_FUNC"
                else:
                    card = "WEIGHT"

                wave = 10 ** hdul[card]["LOGLAM"].read()
                var_lss = hdul[card]["VAR_LSS"].read()

        ax.plot(
            wave,
            var_lss,
            **plot_kwargs,
        )

        ax.set_xlabel(r"$\lambda \, [\AA]$")
        ax.set_ylabel(r"$\sigma^2_{\rm LSS}$")

        return wave, var_lss

    @staticmethod
    def fudge(
        bookkeeper: Optional[Bookkeeper] = None,
        region: Optional[str] = None,
        calib_step: Optional[int] = None,
        attributes_file: Optional[Path | str] = None,
        ax: Optional[Axes] = None,
        plot_kwargs: Dict = dict(),
    ) -> Tuple[wave_grid, wave_grid]:
        if ax is None:
            fig, ax = plt.subplots()

        if isinstance(bookkeeper, Bookkeeper):
            attributes_file = bookkeeper.paths.delta_attributes_file(
                region=region, calib_step=calib_step
            )

        with fitsio.FITS(attributes_file) as hdul:
            if "VAR_FUNC" in hdul:
                card = "VAR_FUNC"
            else:
                card = "WEIGHT"

            lambda_ = 10 ** hdul[card]["LOGLAM"].read()
            fudge = hdul[card]["FUDGE"].read()

        ax.plot(
            lambda_,
            fudge,
            **plot_kwargs,
        )

        ax.set_xlabel(r"$\lambda \, [\AA]$")
        ax.set_ylabel(r"$\epsilon$")

        return lambda_, fudge

    @staticmethod
    def stack(
        bookkeeper: Optional[Bookkeeper] = None,
        region: Optional[str] = None,
        calib_step: Optional[int] = None,
        attributes_file: Optional[Path | str] = None,
        rebin: Optional[int] = None,
        ax: Optional[Axes] = None,
        plot_kwargs: Dict = dict(),
        use_weights: bool = False,
        offset: float = 0,
    ) -> Tuple[wave_grid, wave_grid]:
        if ax is None:
            fig, ax = plt.subplots()

        if isinstance(bookkeeper, Bookkeeper):
            attributes_file = bookkeeper.paths.delta_attributes_file(
                region=region, calib_step=calib_step
            )

        with fitsio.FITS(attributes_file) as hdul:
            if "STATS" in hdul:
                # Raw deltas
                lambda_ = hdul["STATS"]["LAMBDA"].read()
                stack = hdul["STATS"]["MEANFLUX"].read()

                if use_weights:
                    weights = hdul["STATS"]["WEIGHTS"].read()
            else:
                lambda_ = 10 ** hdul["STACK_DELTAS"]["loglam"].read()
                stack = hdul["STACK_DELTAS"]["stack"].read()

                if use_weights:
                    weights = hdul["STACK_DELTAS"]["weight"].read()

        if rebin is not None:
            # Repeat last values till having a number of data multiple of rebin.
            lambda_ = np.concatenate(
                (lambda_, lambda_[-1] * np.ones(rebin - len(lambda_) % rebin))
            )
            stack = np.concatenate(
                (stack, stack[-1] * np.ones(rebin - len(stack) % rebin))
            )

            lambda_ = np.mean(lambda_.reshape(-1, rebin), axis=1)
            stack = np.mean(stack.reshape(-1, rebin), axis=1)

            if use_weights:
                weights = np.concatenate(
                    (weights, np.full(rebin - len(weights) % rebin, weights[-1]))
                )
                weights = np.mean(weights.reshape(-1, rebin), axis=1)

        if use_weights:
            stack *= weights

        ax.plot(lambda_, stack + offset, **plot_kwargs)
        ax.set_xlabel(r"$\lambda \, [\AA]$")
        ax.set_ylabel(r"$\overline{1 + \delta_q(\lambda)}$")

        return lambda_, stack

    @staticmethod
    def num_pixels(
        bookkeeper: Optional[Bookkeeper] = None,
        region: Optional[str] = None,
        calib_step: Optional[int] = None,
        attributes_file: Optional[Path | str] = None,
        ax: Optional[Axes] = None,
        plot_kwargs: Dict = dict(),
    ) -> Tuple[wave_grid, wave_grid_int]:
        if ax is None:
            fig, ax = plt.subplots()

        if isinstance(bookkeeper, Bookkeeper):
            attributes_file = bookkeeper.paths.delta_attributes_file(
                region=region, calib_step=calib_step
            )

        with fitsio.FITS(attributes_file) as hdul:
            lambda_ = 10 ** hdul["VAR_FUNC"]["loglam"].read()
            num_pixels = hdul["VAR_FUNC"]["num_pixels"].read()

        ax.plot(
            lambda_,
            num_pixels,
            **plot_kwargs,
        )
        ax.set_xlabel(r"$\lambda \, [\AA]$")
        ax.set_ylabel(r"# pixels")

        return lambda_, num_pixels

    @staticmethod
    def valid_fit(
        bookkeeper: Optional[Bookkeeper] = None,
        region: Optional[str] = None,
        calib_step: Optional[int] = None,
        attributes_file: Optional[Path | str] = None,
        ax: Optional[Axes] = None,
        plot_kwargs: Dict = dict(),
    ) -> None:
        if ax is None:
            fig, ax = plt.subplots()

        if isinstance(bookkeeper, Bookkeeper):
            attributes_file = bookkeeper.paths.delta_attributes_file(
                region=region, calib_step=calib_step
            )

        with fitsio.FITS(attributes_file) as hdul:
            ax.plot(
                10 ** hdul["VAR_FUNC"]["loglam"].read(),
                hdul["VAR_FUNC"]["valid_fit"].read(),
                **plot_kwargs,
            )

    @staticmethod
    def mean_cont(
        bookkeeper: Optional[Bookkeeper] = None,
        region: Optional[str] = None,
        calib_step: Optional[int] = None,
        attributes_file: Optional[Path | str] = None,
        ax: Optional[Axes] = None,
        plot_kwargs: Dict = dict(),
    ) -> Tuple[wave_grid_rf, wave_grid_rf]:
        if ax is None:
            fig, ax = plt.subplots()

        if isinstance(bookkeeper, Bookkeeper):
            attributes_file = bookkeeper.paths.delta_attributes_file(
                region=region, calib_step=calib_step
            )

        with fitsio.FITS(attributes_file) as hdul:
            wave = 10 ** hdul["CONT"]["loglam_rest"].read()
            mean_cont = hdul["CONT"]["mean_cont"].read()

        ax.plot(
            wave,
            mean_cont,
            **plot_kwargs,
        )

        return wave, mean_cont

    @staticmethod
    def line_masking(
        mask_file: Path
        | str = Path(str(os.getenv("pr")))
        / "Continuum_fitting/config_files/sharp-lines-mask.txt",
        ax: Optional[Axes] = None,
        plot_kwargs: Dict = dict(),
        use_labels: bool = False,
        lambda_lim: Optional[List[float]] = None,
        standard_width: Optional[float] = None,
    ) -> None:
        if ax is None:
            fig, ax = plt.subplots()

        with open(mask_file) as f:
            lines = f.readlines()
        colors = dict(K="green", H="pink", Na="olive", sky="cyan")

        if use_labels:
            label_used = dict()
            for key in colors.keys():
                label_used[key] = False

        for line in lines:
            if line[0] == "#":
                continue

            label = line.split(" ")[0]
            if use_labels:
                label_to_use = label if not label_used[label] else None
                label_used[label] = True
            else:
                label_to_use = None

            lambda_min = float(line.split(" ")[-3])
            lambda_max = float(line.split(" ")[-2])

            if lambda_lim is not None:
                if (lambda_max < lambda_lim[0]) or (lambda_min > lambda_lim[1]):
                    continue

            if standard_width:
                mean_lambda = 0.5 * (lambda_max + lambda_min)
                lambda_min = mean_lambda - 0.5 * standard_width
                lambda_max = mean_lambda + 0.5 * standard_width

            ax.axvspan(
                lambda_min,
                lambda_max,
                facecolor=colors[label],
                label=label_to_use,
                alpha=0.4,
            )

    @staticmethod
    def plot_flux(
        bookkeeper: Bookkeeper,
        los_id: int,
        plot_kwargs: Dict = dict(),
        ax: Optional[Axes] = None,
        z: Optional[float] = None,
    ) -> None:
        if ax is None:
            fig, ax = plt.subplots()

        flux_data = get_spectra_from_los_id(
            los_id,
            catalog=bookkeeper.paths.catalog,
            input_healpix=bookkeeper.paths.healpix_data,
        )

        if z is not None:
            grid = flux_data[0] / (1 + z)
        else:
            grid = flux_data[0]

        ax.plot(
            grid,
            flux_data[1],
            **plot_kwargs,
        )

    @staticmethod
    def plot_cont(
        los_id: int,
        attrs_file: Path | str,
        ax: Optional[Axes] = None,
        region: str = "lya",
        z: Optional[float] = None,
        **kwargs: Any,
    ) -> Tuple[wave_grid_rf, wave_grid_rf]:
        """Helper function to compute continuum

        Arguments:
        ----------
        ax: matplotlib axis where to draw the pot.
        los_id: Line of sight id of quasar.
        attrs_file: Path to delta attributes file.
        region: Region where to compute the continuum.
        z: Redshfit to plot in observed wavelength.
        """
        if ax is None:
            fig, ax = plt.subplots()

        lambda_rest, cont = compute_cont(los_id, attrs_file, region)

        if z is None:
            ax.plot(lambda_rest, cont, **kwargs)
            return lambda_rest, cont
        else:
            ax.plot(lambda_rest * (1 + z), cont, **kwargs)
            return lambda_rest * (1 + z), cont
